#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

from __future__ import annotations

import datetime
import functools
import json
import pathlib
import re
from copy import deepcopy
from typing import List

from PySide6.QtCore import (QPoint, Qt, )
from PySide6.QtGui import (QAction, QIcon, )
from PySide6.QtWidgets import (QApplication, QDialog, QGridLayout, QMenu, QToolBar, )

import revedaEditor.backend.dataDefinitions as ddef
import revedaEditor.backend.libBackEnd as libb
import revedaEditor.backend.libraryMethods as libm
import revedaEditor.backend.libraryModelView as lmview
import revedaEditor.common.shapes as shp  # import the shapes
import revedaEditor.fileio.symbolEncoder as symenc
import revedaEditor.gui.editorViews as edv
import revedaEditor.gui.editorWindow as edw
import revedaEditor.gui.fileDialogues as fd
import revedaEditor.gui.propertyDialogues as pdlg
import revedaEditor.gui.toolsDialogues as tdlg
import revedaEditor.scenes.schematicScene as schscn
from revedaEditor.gui.editorFactory import EditorFactory


class schematicEditor(edw.editorWindow):
    MAJOR_GRID_DEFAULT = 20
    SNAP_GRID_DEFAULT = 10

    def __init__(self, viewItem: libb.viewItem, libraryDict: dict, libraryView) -> None:
        super().__init__(viewItem, libraryDict, libraryView)
        self.setWindowTitle(f"Schematic Editor - {self.cellName} - {self.viewName}")
        self.setWindowIcon(QIcon(":/icons/edLayer-shape.png"))
        self.configDict = dict()
        # self.processedCells = set()  # cells included in config view
        self.symbolChooser = None
        self.majorGrid = self.MAJOR_GRID_DEFAULT  # dot/line grid spacing
        self.snapGrid = self.SNAP_GRID_DEFAULT  # snapping grid size
        self.snapTuple = (self.snapGrid, self.snapGrid)
        self.symbolViews = [
            "symbol"]  # only symbol can be instantiated in the schematic window.
        self._schematicContextMenu()

    def init_UI(self):
        self.resize(1600, 800)
        self._createActions()
        self._createMenuBar()
        self._createToolBars()
        self._addActions()
        self._createTriggers()
        self._createShortcuts()
        self.centralW = schematicContainer(self)
        self.setCentralWidget(self.centralW)

    def __repr__(self):
        return f'schematicEditor({self.libName}-{self.cellName}-{self.viewName})'

    def _createActions(self):
        super()._createActions()
        self.netNameAction = QAction("Net Name", self)
        self.netNameAction.setToolTip("Set Net Name")
        self.netNameAction.setShortcut(Qt.Key.Key_L)
        self.hilightNetAction = QAction("Highlight Net", self)
        self.hilightNetAction.setToolTip("Highlight Selected Net Connections")
        self.hilightNetAction.setCheckable(True)
        self.renumberInstanceAction = QAction("Renumber Instances", self)
        self.renumberInstanceAction.setToolTip("Renumber Instances")
        simulationIcon = QIcon("icons/application-run.png")
        self.simulateAction = QAction(simulationIcon, "Revolution EDA SAE...", self)
        self.findRelatedEditors = QAction("Find Related Editors", self)

    def _addActions(self):
        super()._addActions()
        # edit menu
        self.menuEdit.addAction(self.netNameAction)
        self.menuEdit.removeAction(self.stretchAction)

        self.propertyMenu = self.menuEdit.addMenu("Properties")
        self.propertyMenu.addAction(self.objPropAction)

        # hierarchy submenu
        self.hierMenu = self.menuEdit.addMenu("Hierarchy")
        self.hierMenu.addAction(self.goUpAction)
        self.hierMenu.addAction(self.goDownAction)

        # create menu
        self.menuCreate.addAction(self.createInstAction)
        self.menuCreate.addAction(self.createNetAction)
        self.menuCreate.addAction(self.createBusAction)
        self.menuCreate.addAction(self.createPinAction)
        self.menuCreate.addAction(self.createTextAction)
        self.menuCreate.addAction(self.createSymbolAction)

        # check menu
        self.menuCheck.addAction(self.viewErrorsAction)
        self.menuCheck.addAction(self.deleteErrorsAction)

        # tools menu
        self.menuTools.addAction(self.hilightNetAction)
        self.menuTools.addAction(self.renumberInstanceAction)
        self.menuTools.addAction(self.findRelatedEditors)
        # utilities Menu
        self.selectMenu = self.menuUtilities.addMenu("Selection")
        self.selectMenu.addAction(self.selectDeviceAction)
        self.selectMenu.addAction(self.selectNetAction)
        self.selectMenu.addAction(self.selectPinAction)
        self.selectMenu.addSeparator()
        self.selectMenu.addAction(self.removeSelectFilterAction)
        self.simulationMenu = QMenu("&Simulation")
        # help menu
        self.simulationMenu.addAction(self.netlistAction)
        self.editorMenuBar.insertMenu(self.menuHelp.menuAction(), self.simulationMenu)
        # self.menuHelp = self.editorMenuBar.addMenu("&Help")
        if hasattr(self._app, 'pluginsObj') and hasattr(self._app.pluginsObj, 'applyPluginMenus'):
            self._app.pluginsObj.applyPluginMenus(self)

    def _createTriggers(self):
        super()._createTriggers()

        self.createNetAction.triggered.connect(self.createNetClick)
        self.createBusAction.triggered.connect(self.createBusClick)
        self.createInstAction.triggered.connect(self.createInstClick)
        self.createPinAction.triggered.connect(self.createPinClick)
        self.createTextAction.triggered.connect(self.createNoteClick)
        self.createSymbolAction.triggered.connect(self.createSymbolClick)

        self.objPropAction.triggered.connect(self.objPropClick)
        self.netlistAction.triggered.connect(self.createNetlistClick)
        self.ignoreAction.triggered.connect(self.ignoreClick)
        self.goDownAction.triggered.connect(self.goDownClick)
        self.viewErrorsAction.triggered.connect(self.checkErrorsClick)
        self.deleteErrorsAction.triggered.connect(self.deleteErrorsClick)

        self.hilightNetAction.triggered.connect(self.hilightNetClick)
        self.netNameAction.triggered.connect(self.createNetNameClick)
        self.selectDeviceAction.triggered.connect(self.selectDeviceClick)
        self.selectNetAction.triggered.connect(self.selectNetClick)
        self.selectPinAction.triggered.connect(self.selectPinClick)
        self.removeSelectFilterAction.triggered.connect(self.removeSelectFilterClick)
        self.renumberInstanceAction.triggered.connect(self.renumberInstanceClick)
        self.findRelatedEditors.triggered.connect(self.findRelatedEditorsClick)

    def _createToolBars(self):
        super()._createToolBars()
        self.toolbar.addAction(self.objPropAction)
        self.toolbar.addAction(self.viewPropAction)

        self.schematicToolbar = QToolBar("Schematic Toolbar", self)
        self.addToolBar(self.schematicToolbar)
        self.schematicToolbar.addAction(self.createInstAction)
        self.schematicToolbar.addAction(self.createNetAction)
        self.schematicToolbar.addAction(self.createBusAction)
        self.schematicToolbar.addAction(self.createPinAction)
        # self.schematicToolbar.addAction(self.createLabelAction)
        self.schematicToolbar.addAction(self.createSymbolAction)
        self.schematicToolbar.addSeparator()
        self.schematicToolbar.addAction(self.viewCheckAction)
        self.schematicToolbar.addSeparator()
        self.schematicToolbar.addAction(self.goDownAction)
        self.schematicToolbar.addSeparator()
        self.schematicToolbar.addAction(self.selectDeviceAction)
        self.schematicToolbar.addAction(self.selectNetAction)
        self.schematicToolbar.addAction(self.selectPinAction)
        self.schematicToolbar.addAction(self.removeSelectFilterAction)

    def _schematicContextMenu(self):
        super()._editorContextMenu()
        self.centralW.scene.itemContextMenu.addAction(self.ignoreAction)
        self.centralW.scene.itemContextMenu.addAction(self.goDownAction)

    def _createShortcuts(self):
        super()._createShortcuts()
        self.createInstAction.setShortcut(Qt.Key.Key_I)
        self.createNetAction.setShortcut(Qt.Key.Key_W)
        self.createPinAction.setShortcut(Qt.Key.Key_P)
        self.goDownAction.setShortcut("Shift+E")
        # unset stretch shortcut for schematic editor
        self.stretchAction.setShortcut("")

    def createNetClick(self, s):
        self.centralW.scene.editModes.setMode("drawWire")

    def createBusClick(self, s):
        self.centralW.scene.editModes.setMode("drawBus")

    def createInstClick(self, s):
        # create a designLibrariesView
        libraryModel = lmview.symbolViewsModel(self.libraryDict, self.symbolViews)
        if self.symbolChooser is None:
            self.symbolChooser = fd.selectCellViewDialog(self, libraryModel)
            self.symbolChooser.show()
        else:
            self.symbolChooser.raise_()
        if self.symbolChooser.exec() == QDialog.DialogCode.Accepted:
            instanceTuple = ddef.viewNameTuple(self.symbolChooser.libNamesCB.currentText(),
                                               self.symbolChooser.cellCB.currentText(),
                                               self.symbolChooser.viewCB.currentText(), )

            self.centralW.scene.newInstanceTuple = instanceTuple

            self.centralW.scene.editModes.setMode("addInstance")

    def createPinClick(self, s):
        createPinDlg = pdlg.createSchematicPinDialog(self)
        if createPinDlg.exec() == QDialog.DialogCode.Accepted:
            self.centralW.scene.pinName = createPinDlg.pinName.text()
            self.centralW.scene.pinType = createPinDlg.pinType.currentText()
            self.centralW.scene.pinDir = createPinDlg.pinDir.currentText()
            self.centralW.scene.editModes.setMode("drawPin")

    def createNoteClick(self, s):
        textDlg = pdlg.noteTextEdit(self)
        if textDlg.exec() == QDialog.DialogCode.Accepted:
            noteText = textDlg.plainTextEdit.toPlainText()
            noteFontFamily = textDlg.familyCB.currentText()
            noteFontSize = textDlg.fontsizeCB.currentText()
            noteFontStyle = textDlg.fontStyleCB.currentText()
            noteAlign = textDlg.textAlignmCB.currentText()
            noteOrient = textDlg.textOrientCB.currentText()
            self.centralW.scene.textTuple = (noteText, noteFontFamily, noteFontStyle,
                                             noteFontSize, noteAlign, noteOrient,)
            self.centralW.scene.editModes.setMode("drawText")

    def createSymbolClick(self, s):
        self.createSymbol()

    def objPropClick(self, s):
        self.centralW.scene.editModes.setMode("selectItem")
        self.centralW.scene.viewObjProperties()

    def renumberInstanceClick(self, s):
        self.centralW.scene.renumberInstances()

    def checkSaveCell(self):
        self.centralW.scene.nameSceneNets()
        self.centralW.scene.saveSchematic(self.file)
        if self.parentEditor:
            self.parentEditor.centralW.scene.reloadScene()

    def saveCell(self):
        self.centralW.scene.saveSchematic(self.file)

    def loadSchematic(self):
        try:
            self.logger.info(f'Loading schematic from {self.cellName} - {self.viewName}')
            self.centralW.scene.loadDesign(self.file)
            viewNameTuple = ddef.viewNameTuple(self.libItem.libraryName, self.cellItem.cellName,
                                               self.viewName)
            self.appMainW.openViews[viewNameTuple] = self
        except Exception as e:
            self.logger.error(f"Error during loading schematic for {self.cellName}: {e}")

    def createConfigView(self, configItem: libb.viewItem, newConfigDict: dict,
                         processedCells: set):
        sceneSymbolSet = self.centralW.scene.findSceneSymbolSet()
        for item in sceneSymbolSet:
            libItem = libm.getLibItem(self.libraryView.libraryModel, item.libraryName)
            cellItem = libm.getCellItem(libItem, item.cellName)
            viewItems = [cellItem.child(row) for row in range(cellItem.rowCount())]
            viewNames = [viewItem.viewName for viewItem in viewItems]
            netlistableViews = [viewItemName for viewItemName in self.switchViewList if
                                viewItemName in viewNames]
            itemSwitchViewList = deepcopy(netlistableViews)
            viewDict = dict(zip(viewNames, viewItems))
            itemCellTuple = ddef.cellTuple(libItem.libraryName, cellItem.cellName)
            if itemCellTuple not in processedCells:
                if cellLine := newConfigDict.get(cellItem.cellName):
                    netlistableViews = [cellLine[1]]
                for viewName in netlistableViews:
                    match viewDict[viewName].viewType:
                        case "schematic":
                            newConfigDict[cellItem.cellName] = [libItem.libraryName,
                                                                viewName,
                                                                itemSwitchViewList, ]
                            schematicObj = schematicEditor(viewDict[viewName],
                                                           self.libraryDict,
                                                           self.libraryView, )
                            schematicObj.loadSchematic()
                            schematicObj.createConfigView(configItem, newConfigDict,
                                                          processedCells)
                            break
                        case _:
                            newConfigDict[cellItem.cellName] = [libItem.libraryName,
                                                                viewName,
                                                                itemSwitchViewList, ]
                            break
                processedCells.add(itemCellTuple)

    def closeEvent(self, event):
        try:
            self.centralW.scene.saveSchematic(self.file)
            cellViewNameTuple = ddef.viewNameTuple(self.libName, self.cellName,
                                                   self.viewName)
            self.appMainW.openViews.pop(cellViewNameTuple, None)
        except Exception as e:
            self.appMainW.logger.error(f"Error in closing schematic editor window"
                                       f":{self.cellName}-{self.viewName}:{e}")
        finally:
            event.accept()
            super().closeEvent(event)

    def createNetlistClick(self, s):
        dlg = fd.netlistExportDialogue(self)
        dlg.libNameEdit.setText(self.libName)
        dlg.cellNameEdit.setText(self.cellName)
        netlistableViews = self._getNetlistableViews()
        dlg.viewNameCombo.addItems(netlistableViews)

        if hasattr(self.appMainW, "outputPrefixPath"):
            dlg.netlistDirEdit.setText(str(self.appMainW.outputPrefixPath))

        if dlg.exec() == QDialog.DialogCode.Accepted:
            self._startNetlisting(dlg)

    def _getNetlistableViews(self):
        views = [self.viewItem.viewName]
        config_items = [self.cellItem.child(row) for row in range(self.cellItem.rowCount())
                        if self.cellItem.child(row).viewType == "config"]

        for item in config_items:
            with item.data(Qt.ItemDataRole.UserRole + 2).open(mode="r") as f:
                config = json.load(f)
                if config[1]["reference"] == self.viewItem.viewName:
                    views.append(item.viewName)

        return views

    def _startNetlisting(self, dlg: fd.netlistExportDialogue):
        # try:
        self.appMainW.simulationOutputPath = pathlib.Path(dlg.netlistDirEdit.text())
        selectedViewName = dlg.viewNameCombo.currentText()

        self.switchViewList = [item.strip() for item in
                               dlg.switchViewEdit.text().split(",")]
        self.stopViewList = [dlg.stopViewEdit.text().strip()]

        subDirPath = self.appMainW.simulationOutputPath / self.libName / self.cellName / selectedViewName
        subDirPath.mkdir(parents=True, exist_ok=True)

        netlistFilePath = subDirPath / f"{self.cellName}_{selectedViewName}.cir"
        if dlg.topAsSubcktCheckBox.isChecked():
            topSubCkt = True
        else:
            topSubCkt = False

        netlistObj = self.createNetlistObject(self.viewItem, netlistFilePath, topSubCkt)

        if netlistObj:
            # self.runNetlisting(netlistObj, self.appMainW.threadPool)
            with self.measureDuration():
                netlistObj.writeNetlist()

    def createNetlistObject(self, viewItem: libb.viewItem, filePath: pathlib.Path,
                            topSubCkt: bool):
        if viewItem.viewType == "schematic":
            return xyceNetlist(self, filePath, False, topSubCkt)
        elif viewItem.viewType == "config":
            netlistObj = xyceNetlist(self, filePath, True, topSubCkt)
            configItem = libm.findViewItem(self.libraryView.libraryModel, self.libName,
                                           self.cellName, viewItem.viewName)
            if configItem:
                with configItem.data(Qt.ItemDataRole.UserRole + 2).open(mode="r") as f:
                    netlistObj.configDict = json.load(f)[2]
            return netlistObj
        return None

    def goDownClick(self, s):
        self.centralW.scene.goDownHier()

    def ignoreClick(self, s):
        self.centralW.scene.ignoreSymbol()

    def createNetNameClick(self, s):
        dlg = pdlg.netNameDialog(self)
        if dlg.exec() == QDialog.DialogCode.Accepted:
            self.centralW.scene.netNameString = dlg.netNameEdit.text().strip()
            self.centralW.scene.editModes.setMode('nameNet')
        self.messageLine.setText("Select a net to attach the name")

    def hilightNetClick(self, s):
        self.centralW.scene.hilightNets()

    def selectDeviceClick(self):
        self.centralW.scene.selectModes.setMode("selectDevice")
        self.messageLine.setText("Select Only Instances")

    def selectNetClick(self):
        self.centralW.scene.selectModes.setMode("selectNet")
        self.messageLine.setText("Select Only Nets")

    def selectPinClick(self):
        self.centralW.scene.selectModes.setMode("selectPin")
        self.messageLine.setText("Select Only Pins")

    def removeSelectFilterClick(self):
        self.centralW.scene.selectModes.setMode("selectAll")
        self.messageLine.setText("Select All Objects")

    def createSymbol(self) -> None:
        """
        Create a symbol view for a schematic.
        """
        oldSymbolItem = False

        askViewNameDlg = pdlg.symbolNameDialog(self.file.parent, self.cellName, self, )
        if askViewNameDlg.exec() == QDialog.DialogCode.Accepted:
            symbolViewName = askViewNameDlg.symbolViewsCB.currentText()
            if symbolViewName in askViewNameDlg.symbolViewNames:
                oldSymbolItem = True
            if oldSymbolItem:
                deleteSymViewDlg = fd.deleteSymbolDialog(self.cellName, symbolViewName,
                                                         self)
                if deleteSymViewDlg.exec() == QDialog.DialogCode.Accepted:
                    self.generateSymbol(symbolViewName)
            else:
                self.generateSymbol(symbolViewName)

    def _parse_pin_names(self, text: str) -> list[str]:
        """Parse comma-separated pin names, filtering empty strings."""
        return [name.strip() for name in text.split(",") if name.strip()]

    def _categorize_pins(self, schematic_pins: list[shp.schematicPin]) -> dict:
        """Categorize pins by direction."""
        pin_dirs = shp.schematicPin.pinDirs
        return {
            'input': [pin.pinName for pin in schematic_pins if pin.pinDir == pin_dirs[0]],
            'output': [pin.pinName for pin in schematic_pins if pin.pinDir == pin_dirs[1]],
            'inout': [pin.pinName for pin in schematic_pins if pin.pinDir == pin_dirs[2]]}

    def _create_pin_map(self, schematic_pins: list[shp.schematicPin]) -> dict:
        """Create mapping from pin name to pin object."""
        return {pin.pinName: pin for pin in schematic_pins}

    def _draw_pins(self, symbol_scene, pin_names: list[str], locations: list[QPoint],
                   offsets: QPoint, pin_map: dict) -> None:
        """Draw pins and their connecting lines."""
        for name, loc in zip(pin_names, locations):
            symbol_scene.lineDraw(loc, loc + offsets)
            symbol_scene.addItem(pin_map[name].toSymbolPin(loc))

    def generateSymbol(self, symbolViewName: str) -> None:
        # Get schematic pins and categorize them
        schematic_pins = list(self.centralW.scene.findSceneSchemPinsSet())
        pin_categories = self._categorize_pins(schematic_pins)
        pin_map = self._create_pin_map(schematic_pins)

        # Setup dialog with categorized pins
        dlg = pdlg.symbolCreateDialog(self)
        dlg.leftPinsEdit.setText(", ".join(pin_categories['input']))
        dlg.rightPinsEdit.setText(", ".join(pin_categories['output']))
        dlg.topPinsEdit.setText(", ".join(pin_categories['inout']))

        if dlg.exec() != QDialog.DialogCode.Accepted:
            return

        # Create symbol view and editor
        libItem = libm.getLibItem(self.libraryView.libraryModel, self.libName)
        cellItem = libm.getCellItem(libItem, self.cellName)
        symbolViewItem = libb.createCellView(self, symbolViewName, cellItem)

        symbolWindow = EditorFactory.createEditor("symbol", symbolViewItem,
                                                  self.libraryDict, self.libraryView)

        try:
            # Parse pin configurations
            pin_names = {'left': self._parse_pin_names(dlg.leftPinsEdit.text()),
                         'right': self._parse_pin_names(dlg.rightPinsEdit.text()),
                         'top': self._parse_pin_names(dlg.topPinsEdit.text()),
                         'bottom': self._parse_pin_names(dlg.bottomPinsEdit.text())}

            stub_length = int(float(dlg.stubLengthEdit.text().strip()))
            pin_distance = int(float(dlg.pinDistanceEdit.text().strip()))

            # Calculate rectangle dimensions
            rect_x_dim = (max(len(pin_names['top']),
                              len(pin_names['bottom'])) + 1) * pin_distance
            rect_y_dim = (max(len(pin_names['left']),
                              len(pin_names['right'])) + 1) * pin_distance

        except ValueError:
            self.logger.error("Enter valid value")
            return

        # Draw symbol components
        sceneSymbol = symbolWindow.centralW.scene
        sceneSymbol.rectDraw(QPoint(0, 0), QPoint(rect_x_dim, rect_y_dim))

        # Add labels
        sceneSymbol.labelDraw(QPoint(int(0.25 * rect_x_dim), int(0.4 * rect_y_dim)),
                              "[@cellName]", "NLPLabel", "12", "Center", "R0", "Instance")
        sceneSymbol.labelDraw(QPoint(int(rect_x_dim), int(-0.1 * rect_y_dim)),
                              "[@instName]", "NLPLabel", "12", "Center", "R0", "Instance")

        # Calculate pin locations and draw pins
        pin_configs = [(pin_names['left'],
                        [QPoint(-stub_length, (i + 1) * pin_distance) for i in
                         range(len(pin_names['left']))], QPoint(stub_length, 0)),
                       (pin_names['right'],
                        [QPoint(rect_x_dim + stub_length, (i + 1) * pin_distance) for i in
                         range(len(pin_names['right']))], QPoint(-stub_length, 0)),
                       (pin_names['top'],
                        [QPoint((i + 1) * pin_distance, -stub_length) for i in
                         range(len(pin_names['top']))], QPoint(0, stub_length)),
                       (pin_names['bottom'],
                        [QPoint((i + 1) * pin_distance, rect_y_dim + stub_length) for i in
                         range(len(pin_names['bottom']))], QPoint(0, -stub_length))]

        for names, locations, offset in pin_configs:
            self._draw_pins(sceneSymbol, names, locations, offset, pin_map)

        # Add symbol attributes
        all_pin_names = [pin.pinName for pin in schematic_pins]
        sceneSymbol.attributeList = [
            symenc.symbolAttribute("SpiceNetlistLine", "X@instName %pinOrder @cellName"),
            symenc.symbolAttribute("SpectreNetlistLine",
                                   "X@instName  ( %pinOrder ) @cellName"),
            symenc.symbolAttribute("pinOrder", ", ".join(all_pin_names))]

        # Finalize and show
        symbolWindow.checkSaveCell()
        self.libraryView.reworkDesignLibrariesView(self.appMainW.libraryDict)

        openCellViewTuple = ddef.viewNameTuple(self.libName, self.cellName, symbolViewName)
        self.appMainW.openViews[openCellViewTuple] = symbolWindow
        symbolWindow.show()

    def checkErrorsClick(self):
        self.centralW.scene.checkErrors()
        self.logger.info("Schematic Errors checked.")

    def deleteErrorsClick(self):
        self.centralW.scene.deleteErrors()
        self.logger.info("Schematic Errors checked.")

    def findRelated(self, editor, visited, openWindows):
        if editor in visited or editor not in openWindows:
            return
        visited.add(editor)
        if hasattr(editor, 'parentEditor') and editor.parentEditor:
            self.findRelated(editor.parentEditor, visited, openWindows)
        for w in openWindows:
            if hasattr(w, 'parentEditor') and w.parentEditor == editor:
                self.findRelated(w, visited, openWindows)

    def findRelatedEditorsClick(self):
        openWindows = {w for w in QApplication.instance().topLevelWidgets() if
                       isinstance(w, schematicEditor)}
        relatedEditors = set()
        self.findRelated(self, relatedEditors, openWindows)
        dlg = tdlg.findProjectEditors(self)
        relatedEditorNameList = [editor.cellName for editor in relatedEditors]
        dlg.relatedEditorsCB.addItems(relatedEditorNameList)
        dlg.relatedEditorsCB.setCurrentIndex(0)
        if dlg.exec() == QDialog.DialogCode.Accepted:
            relatedEditorName = dlg.relatedEditorsCB.currentText()
            relatedEditor = next((editor for editor in relatedEditors if
                                  editor.cellName == relatedEditorName), None)
            if relatedEditor:
                relatedEditor.show()
                relatedEditor.raise_()
                relatedEditor.activateWindow()


class schematicContainer(edw.editorContainer):
    def __init__(self, parent: schematicEditor):
        super().__init__(parent=parent)
        self.editorWindow = parent
        self.scene = schscn.schematicScene(self)
        self.view = edv.schematicView(self.scene, self)
        self.init_UI()

    def init_UI(self):
        # layout statements, using a grid layout
        gLayout = QGridLayout()
        gLayout.setSpacing(5)
        gLayout.setContentsMargins(0, 0, 0, 0)  # remove outer gaps
        gLayout.addWidget(self.view, 0, 0)
        gLayout.setColumnStretch(0, 5)
        gLayout.setRowStretch(0, 6)
        self.setLayout(gLayout)


class xyceNetlist:
    # Pre-compiled regex to strip dangling parameter assignments (e.g. ` width =`) left
    # after token substitution.  Compiled once at class level avoids re.compile() on every
    # netlist line inside the hot loop.
    _PARAM_RE = re.compile(r'\s+\w+\s*=(?=\s|$)')

    def __init__(self, schematic, filePathObj: pathlib.Path, useConfig: bool = False,
                 topSubckt: bool = False):
        self.filePathObj = filePathObj
        self.schematic = schematic
        self._useConfig = useConfig
        self._scene = self.schematic.centralW.scene
        self.libraryDict = self.schematic.libraryDict
        self.libraryView = self.schematic.libraryView
        self._configDict = {}
        self.topSubckt = topSubckt
        self.libItem = libm.getLibItem(self.schematic.libraryView.libraryModel,
                                       self.schematic.libName, )
        self.cellItem = libm.getCellItem(self.libItem, self.schematic.cellName)
        self.subcircuitDefs = []
        self._switchViewList = schematic.switchViewList
        self._stopViewList = schematic.stopViewList
        self.netlistedViewsSet = set()  # keeps track of netlisted views.
        self.includeLines = set()  # keeps track of include lines.
        self.vamodelLines = set()  # keeps track of vamodel lines.
        self.vahdlLines = set()  # keeps track of *.HDL lines.
        # Caches to avoid repeated Qt model traversals for the same cells.
        self._viewNameCache: dict[tuple, str] = {}  # (libName, cellName) -> netlist view name
        self._cellItemCache: dict[tuple, object] = {}  # (libName, cellName) -> cellItem

    def __repr__(self):
        return f"xyceNetlist(filePathObj={self.filePathObj}, schematic={self.schematic}, useConfig={self._useConfig})"

    @property
    def switchViewList(self) -> List[str]:
        return self._switchViewList

    @switchViewList.setter
    def switchViewList(self, value: List[str]):
        self._switchViewList = value

    @property
    def stopViewList(self) -> List[str]:
        return self._stopViewList

    @stopViewList.setter
    def stopViewList(self, value: List[str]):
        self._stopViewList = value

    @property
    def configDict(self):
        return self._configDict

    @configDict.setter
    def configDict(self, value: dict):
        self._configDict = value

    def writeNetlist(self):
        with self.filePathObj.open(mode="w") as cirFile:
            cirFile.write("*".join(["\n", 80 * "*", "\n", "* Revolution EDA CDL Netlist\n",
                                    f"* Library: {self.schematic.libName}\n",
                                    f"* Top Cell Name: {self.schematic.cellName}\n",
                                    f"* View Name: {self.schematic.viewName}\n",
                                    f"* Date: {datetime.datetime.now()}\n", 80 * "*", "\n",
                                    ".GLOBAL gnd!\n\n", ]))

            # Initialize subcircuit definitions list
            self.subcircuitDefs = []

            # now go down the rabbit hole to track all circuit elements.
            self.recursiveNetlisting(self.schematic, cirFile)

            # Write all subcircuit definitions at the end
            if self.subcircuitDefs:
                cirFile.write("\n* Subcircuit Definitions\n")
                for subcktDef in self.subcircuitDefs:
                    cirFile.write(subcktDef)

            # cirFile.write(".END\n")
            for line in self.includeLines:
                cirFile.write(f"{line}\n")
            for line in self.vamodelLines:
                cirFile.write(f"{line}\n")
            for line in self.vahdlLines:
                cirFile.write(f"{line}\n")

    def collectSubcircuitContent(self, schematic: schematicEditor, content):
        """Collect subcircuit content without writing to file."""
        schematicScene = schematic.centralW.scene
        schematicScene.nameSceneNets()
        sceneSymbolSet = schematicScene.findSceneSymbolSet()
        schematicScene.generatePinNetMap(sceneSymbolSet)
        for elementSymbol in sceneSymbolSet:
            if elementSymbol.symattrs.get("NetlistIgnore") != "1" and (
                    not elementSymbol.netlistIgnore):
                cellItem = self._getCellItem(elementSymbol.libraryName,
                                             elementSymbol.cellName)
                netlistView = self.determineNetlistView(elementSymbol, cellItem)

                if "schematic" in netlistView:
                    lines = self.createXyceSymbolLine(elementSymbol)
                    content.extend(lines if isinstance(lines, list) else [lines])
                    if netlistView not in self._stopViewList:
                        # Check deduplication before loading the schematic to avoid
                        # creating a schematicEditor object for every repeated instance.
                        viewTuple = ddef.viewNameTuple(elementSymbol.libraryName,
                                                       elementSymbol.cellName, netlistView)
                        if viewTuple not in self.netlistedViewsSet:
                            self.netlistedViewsSet.add(viewTuple)
                            schematicItem = libm.getViewItem(cellItem, netlistView)
                            schematicObj = schematicEditor(schematicItem, self.libraryDict,
                                                           self.libraryView)
                            schematicObj.loadSchematic()
                            expandedPinsString = self.expandPinNames(
                                list(elementSymbol.pinNetMap.keys()))
                            subcktContent = []
                            self.collectSubcircuitContent(schematicObj, subcktContent)
                            subcktDef = f".SUBCKT {schematicObj.cellName} {expandedPinsString}\n" + '\n'.join(
                                subcktContent) + "\n.ENDS\n"
                            self.subcircuitDefs.append(subcktDef)
                elif "symbol" in netlistView:
                    lines = self.createXyceSymbolLine(elementSymbol)
                    content.extend(lines if isinstance(lines, list) else [lines])
                elif "spice" in netlistView:
                    lines = self.createSpiceLine(elementSymbol)
                    content.extend(lines if isinstance(lines, list) else [lines])
                elif "veriloga" in netlistView:
                    lines = self.createVerilogaLine(elementSymbol)
                    content.extend(lines if isinstance(lines, list) else [lines])
            elif elementSymbol.netlistIgnore:
                content.append(f"*{elementSymbol.instanceName} is marked to be ignored\n")
            else:
                # NetlistIgnore attribute == "1" but netlistIgnore flag is False
                content.append(
                    f"*{elementSymbol.instanceName} is excluded via NetlistIgnore attribute\n")

    def recursiveNetlisting(self, schematicEdObj: schematicEditor, cirFile):
        """
        Recursively traverse all sub-circuits and netlist them.
        """
        if self.topSubckt:
            viewTuple = ddef.viewNameTuple(schematicEdObj.libName, schematicEdObj.cellName,
                                           schematicEdObj.viewName)
            self.netlistedViewsSet.add(viewTuple)
            schematicPinsSet = schematicEdObj.centralW.scene.findSceneSchemPinsSet()
            pinNames = [pin.pinName for pin in schematicPinsSet]
            expandedPinsString = self.expandPinNames(pinNames)

            subcktContent = []
            self.collectSubcircuitContent(schematicEdObj, subcktContent)
            subcktDef = f"\n.SUBCKT {schematicEdObj.cellName} {expandedPinsString}\n" + '\n'.join(
                subcktContent) + "\n.ENDS\n"
            self.subcircuitDefs.append(subcktDef)
        else:
            schematicScene = schematicEdObj.centralW.scene
            schematicScene.nameSceneNets()  # name all nets in the schematic
            sceneSymbolSet = schematicScene.findSceneSymbolSet()
            schematicScene.generatePinNetMap(sceneSymbolSet)
            for elementSymbol in sceneSymbolSet:
                self.processElementSymbol(elementSymbol, schematicEdObj, cirFile)

    def processElementSymbol(self, elementSymbol, schematic, cirFile):
        if elementSymbol.symattrs.get("NetlistIgnore") != "1" and (
                not elementSymbol.netlistIgnore):
            cellItem = self._getCellItem(elementSymbol.libraryName,
                                         elementSymbol.cellName)
            netlistView = self.determineNetlistView(elementSymbol, cellItem)

            # Create the netlist line for the item.
            self.createItemLine(cirFile, elementSymbol, cellItem, netlistView)
        elif elementSymbol.netlistIgnore:
            cirFile.write(f"*{elementSymbol.instanceName} is marked to be ignored\n")
        else:
            # NetlistIgnore attribute == "1" but netlistIgnore flag is False
            cirFile.write(
                f"*{elementSymbol.instanceName} is excluded via NetlistIgnore attribute\n")

    def _getCellItem(self, libraryName: str, cellName: str) -> libb.cellItem | None:
        """Return the cellItem for (libraryName, cellName), cached to avoid repeated model traversals."""
        key = (libraryName, cellName)
        if key not in self._cellItemCache:
            libItem = libm.getLibItem(self.libraryView.libraryModel, libraryName)
            self._cellItemCache[key] = libm.getCellItem(libItem, cellName)
        return self._cellItemCache[key]

    def determineNetlistView(self, elementSymbol, cellItem) -> str:
        cacheKey = (elementSymbol.libraryName, elementSymbol.cellName)
        if cacheKey in self._viewNameCache:
            return self._viewNameCache[cacheKey]

        viewItems = [cellItem.child(row) for row in range(cellItem.rowCount())]
        viewNames = [view.viewName for view in viewItems]

        if self._useConfig:
            config_entry = self.configDict.get(elementSymbol.cellName)
            result = config_entry[1] if config_entry else "symbol"
        else:
            # Iterate over the switch view list to determine the appropriate netlist view.
            for viewName in self._switchViewList:
                if viewName in viewNames:
                    result = viewName
                    break
            else:
                result = "symbol"

        self._viewNameCache[cacheKey] = result
        return result

    def createItemLine(self, cirFile, elementSymbol: shp.schematicSymbol,
                       cellItem: libb.cellItem, netlistView: str, ):
        if "schematic" in netlistView:
            elementLines = self.createXyceSymbolLine(elementSymbol)
            for line in elementLines:
                cirFile.write(f"{line}\n")

            if netlistView not in self._stopViewList:
                # Build viewTuple from symbol attributes before touching the filesystem;
                # this avoids creating a heavy schematicEditor object for every instance
                # of an already-netlisted cell (the common case in large designs).
                viewTuple = ddef.viewNameTuple(elementSymbol.libraryName,
                                               elementSymbol.cellName, netlistView)
                if viewTuple not in self.netlistedViewsSet:
                    self.netlistedViewsSet.add(viewTuple)
                    schematicItem = libm.getViewItem(cellItem, netlistView)
                    schematicObj = schematicEditor(schematicItem, self.libraryDict,
                                                   self.libraryView)
                    schematicObj.loadSchematic()
                    expandedPinsString = self.expandPinNames(
                        list(elementSymbol.pinNetMap.keys()))
                    subcktContent = []
                    self.collectSubcircuitContent(schematicObj, subcktContent)
                    subcktDef = f"\n.SUBCKT {schematicObj.cellName} {expandedPinsString}\n" + '\n'.join(
                        subcktContent) + "\n.ENDS\n"
                    self.subcircuitDefs.append(subcktDef)
        elif "symbol" in netlistView:
            symbolLines = self.createXyceSymbolLine(elementSymbol)
            for line in symbolLines:
                cirFile.write(f"{line}\n")
        elif "spice" in netlistView:
            spiceLines = self.createSpiceLine(elementSymbol)
            for line in spiceLines:
                cirFile.write(f"{line}\n")
        elif "veriloga" in netlistView:
            verilogaLines = self.createVerilogaLine(elementSymbol)
            for line in verilogaLines:
                cirFile.write(f"{line}\n")

    def _createNetlistLine(self, elementSymbol: shp.schematicSymbol, netlistLineKey: str) -> \
            list[str]:
        """Shared netlist line creation logic."""
        try:
            instNameLabel = elementSymbol.labels.get('@instName')
            # print(f"Creating netlist line for instance: {instNameLabel.labelValue if instNameLabel else 'N/A'}")
            if not instNameLabel:
                return []

            baseInstName, arrayTuple = self.parseArrayNotation(
                instNameLabel.labelValue.strip())
            arrayStep = -1 if arrayTuple[0] > arrayTuple[1] else 1
            arraySize = abs(arrayTuple[0] - arrayTuple[1]) + 1

            baseNetlistLine = elementSymbol.symattrs[netlistLineKey].strip()
            instNameToken = instNameLabel.labelName
            symbolLines = []

            # Pre-compute substitution lists once; avoids repeated dict iteration
            # for every element in an arrayed instance (e.g. inst<0:99>).
            attr_replacements = [(f"%{a}", v) for a, v in elementSymbol.symattrs.items()]
            label_replacements = [(lbl.labelName, lbl.labelValue)
                                  for lbl in elementSymbol.labels.values()]

            def processLine(line, netsList):
                line = line.replace("%pinOrder", netsList)
                for token, value in attr_replacements:
                    line = line.replace(token, value)
                return xyceNetlist._PARAM_RE.sub('', line)

            def createInstanceLine(instanceName):
                line = baseNetlistLine.replace(instNameToken, instanceName)
                for labelName, labelValue in label_replacements:
                    line = line.replace(labelName, labelValue)
                return line

            def expandNet(netName):
                baseName, netTuple = self.parseArrayNotation(netName)
                if netTuple[0] == netTuple[1] == -1:
                    return [baseName]
                netStep = 1 if netTuple[1] >= netTuple[0] else -1
                return [f"{baseName}<{i}>" for i in
                        range(netTuple[0], netTuple[1] + netStep, netStep)]

            # Expand nets per pin
            expandedPinNets = [expandNet(netName) for netName in elementSymbol.pinNetMap.values()]

            # Generate instance lines
            if arraySize == 1:
                # Scalar instance logic
                flatNetsList = " ".join([nets[0] for nets in expandedPinNets])
                symbolLines.append(processLine(createInstanceLine(baseInstName), flatNetsList))
            else:
                # Array instance logic
                arrayIndices = list(range(arrayTuple[0], arrayTuple[1] + arrayStep, arrayStep))
                
                for j, i in enumerate(arrayIndices):
                    instanceNets = []
                    for nets in expandedPinNets:
                        if len(nets) == arraySize:
                            instanceNets.append(nets[j])  # 1-to-1 matching across array width
                        elif len(nets) == 1:
                            instanceNets.append(nets[0])  # Scalar broadcasted to all array nodes
                        else:
                            # Log a connection width mismatch warning if len is neither 1 nor arraySize
                            self._scene.logger.warning(
                                f"Net connection width mismatch for {elementSymbol.instanceName}: "
                                f"expected 1 or {arraySize}, got {len(nets)}. Falling back to element 0."
                            )
                            instanceNets.append(nets[0])
                            
                    specificNetsList = " ".join(instanceNets)
                    symbolLines.append(
                        processLine(createInstanceLine(f"{baseInstName}<{i}>"), specificNetsList))

            return symbolLines


        except Exception as e:
            self._scene.logger.error(
                f"Error creating netlist line for {elementSymbol.instanceName}: {e}")
            return [
                f"*Netlist line is not defined for symbol of {elementSymbol.instanceName}\n"]

    def createXyceSymbolLine(self, elementSymbol: shp.schematicSymbol) -> list[str]:
        return self._createNetlistLine(elementSymbol, "SpiceNetlistLine")

    def createSpiceLine(self, elementSymbol: shp.schematicSymbol) -> list[str]:
        try:
            spiceLines = self.createXyceSymbolLine(elementSymbol)
            incFileName = elementSymbol.symattrs.get("incLine", "").strip()
            if incFileName:
                cellItem = self._getCellItem(elementSymbol.libraryName, elementSymbol.cellName)
                cellPath = cellItem.data(Qt.ItemDataRole.UserRole + 2)
                incFilePath = pathlib.Path(cellPath) / incFileName
                self.includeLines.add(f'.INC "{incFilePath}"')
            else:
                self.includeLines.add(f"* no include line found for {elementSymbol.cellName}")
            return spiceLines
        except Exception as e:
            self._scene.logger.error(f"Spice subckt netlist error: {e}")
            return [f"*Netlist line is not defined for symbol of {elementSymbol.instanceName}\n"]

    def createVerilogaLine(self, elementSymbol) -> list[str]:
        try:
            symbolLines = self._createNetlistLine(elementSymbol, "SpiceNetlistLine")
            self.vamodelLines.add(elementSymbol.symattrs.get("vaModelLine",
                                                             "* no model line is found for {item.cellName}").strip())
            self.vahdlLines.add(elementSymbol.symattrs.get("vaHDLLine",
                                                           "* no hdl line is found for {item.cellName}").strip())
            return symbolLines
        except Exception as e:
            self._scene.logger.error(e)
            return [f"*Netlist line is not defined for symbol of {elementSymbol.instanceName}\n"]

    @staticmethod
    @functools.lru_cache(maxsize=None)
    def parseArrayNotation(name: str) -> tuple[str, tuple[int, int]]:
        """
        Parse net/instance array notation like 'name<0:5>' into base name and index range.
        Also handles single instance/net notation like 'name<0>' or 'name<1>'.

        Args:
        name (str): The net name with optional bus notation.

        Returns:
        tuple[str, tuple[int, int]]: A tuple containing the base name and a tuple of start and end indices.
        """
        # Check if the name does not contain bus notation
        if '<' not in name or '>' not in name:
            return name, (-1, -1)

        baseName = name.split('<')[0]  # Extract the base name before '<'
        indexRange = name.split('<')[1].split('>')[0]  # Extract the content inside '<>'

        # Check if it's a single index (e.g., 'name<0>')
        if ':' not in indexRange:
            singleIndex = int(indexRange)
            return baseName, (singleIndex, singleIndex)

        # Handle range notation (e.g., 'name<0:5>')
        start, end = map(int, indexRange.split(':'))
        return baseName, (start, end)

    @staticmethod
    def expandPinNames(pinNamesList: List[str]) -> str:
        expandedPinNameList = []
        for pinName in pinNamesList:
            pinBaseName, pinTuple = xyceNetlist.parseArrayNotation(pinName)
            if pinTuple[0] == pinTuple[1] == -1:
                expandedPinNameList.append(pinBaseName)
            else:
                pinStep = 1 if pinTuple[1] >= pinTuple[0] else -1
                for i in range(pinTuple[0], pinTuple[1] + pinStep):
                    expandedPinNameList.append(f'{pinBaseName}<{i}>')
        return ' '.join(expandedPinNameList)



#     "Commons Clause" License Condition v1.0
#    #
#     The Software is provided to you by the Licensor under the License, as defined
#     below, subject to the following condition.
#  #
#     Without limiting other conditions in the License, the grant of rights under the
#     License will not include, and the License does not grant to you, the right to
#     Sell the Software.
#  #
#     For purposes of the foregoing, "Sell" means practicing any or all of the rights
#     granted to you under the License to provide to third parties, for a fee or other
#     consideration (including without limitation fees for hosting) a product or service whose value
#     derives, entirely or substantially, from the functionality of the Software. Any
#     license notice or attribution required by the License must also include this
#     Commons Clause License Condition notice.
#  #
#    Add-ons and extensions developed for this software may be distributed
#    under their own separate licenses.
#  #
#     Software: Revolution EDA
#     License: Mozilla Public License 2.0
#     Licensor: Revolution Semiconductor (Registered in the Netherlands)

from typing import List, Dict, Any

from PySide6.QtCore import (QAbstractTableModel, Qt, QModelIndex, QPersistentModelIndex,
                            Signal, QPoint, QRect)
from PySide6.QtGui import QFont
from PySide6.QtWidgets import (QHeaderView, QTableView, QMenu)
from revedaEditor.backend.pdkLoader import importPDKModule

process = importPDKModule("process")



class DRCTableModel(QAbstractTableModel):
    def __init__(self, violations: List[Dict[str, Any]],
                 categories: Dict[str, str]):
        super().__init__()
        self._data = violations

        self._categories = categories
        self._headers = ['#', 'Category', 'Description', 'Cell', 'Visited',
                         'Multiplicity', 'Points']

    def rowCount(self, parent: QModelIndex | QPersistentModelIndex = QModelIndex()) -> int:
        return len(self._data)

    def columnCount(self, parent: QModelIndex | QPersistentModelIndex = QModelIndex()):
        return len(self._headers)

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid() or role != Qt.DisplayRole:
            return None

        row = self._data[index.row()]

        col = index.column()

        # Handle case where row might be a string instead of dict
        if isinstance(row, str):
            return str(index.row() + 1) if col == 0 else (row if col == 1 else "")

        if col == 0:
            return str(index.row() + 1)
        elif col == 1:
            return row.get('category', '')
        elif col == 2:
            return self._categories.get(row.get('category', ''))
        elif col == 3:
            return row.get('cell', '')
        elif col == 4:
            return str(row.get('visited', ''))
        elif col == 5:
            return str(row.get('multiplicity', ''))
        elif col == 6:
            return str(row.get('points', ''))

        return None

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Orientation.Horizontal:
            if role == Qt.DisplayRole:
                return self._headers[section]
            elif role == Qt.FontRole:
                font = QFont()
                font.setBold(True)
                return font
        return None

    def getPolygons(self, row):
        return self._data[row]['polygons']
    

    def markVisited(self, row):
        if 0 <= row < len(self._data):
            self._data[row]['visited'] = True
            index = self.index(row, 4)  # Column 4 is 'Visited'
            self.dataChanged.emit(index, index)


class DRCTableView(QTableView):
    polygonSelected = Signal(list)  # Signal to emit selected polygons
    zoomToRect = Signal(QRect) # Signal to emit polygon to be zoomed.
    def __init__(self, data, categories):
        super().__init__()
        self.drcOutputsModel = DRCTableModel(data, categories)
        self.setModel(self.drcOutputsModel)
        self.selectionModel().currentRowChanged.connect(self.onRowChanged)
        self.header = self.horizontalHeader()
        self.header.setSectionResizeMode(0, QHeaderView.ResizeToContents)
        self.header.setSectionResizeMode(1, QHeaderView.ResizeToContents)
        self.header.setSectionResizeMode(2, QHeaderView.Stretch)
        self.header.setSectionResizeMode(3, QHeaderView.ResizeToContents)
        self.header.setSectionResizeMode(4, QHeaderView.ResizeToContents)
        self.header.setSectionResizeMode(5, QHeaderView.ResizeToContents)
        self.header.setSectionResizeMode(6, QHeaderView.ResizeToContents)
        self.header.setMaximumSectionSize(200)
        self.header.setStretchLastSection(False)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._onContextMenuRequested)

    def onRowChanged(self, current, previous):
        if current.isValid():
            row = current.row()
            self.drcOutputsModel.markVisited(row)
            polygons = self.drcOutputsModel.getPolygons(row)
            self.polygonSelected.emit(polygons)

    def _onContextMenuRequested(self, pos: QPoint):
        index = self.indexAt(pos)
        if not index.isValid():
            return
        row = index.row()
        menu = QMenu(self)
        # Example actions (customize as needed)
        copy_points = menu.addAction("Zoom To Error")
        copy_points.triggered.connect(lambda: self._zoomToError(row))
        # menu.addAction("Other action...")
        menu.exec(self.viewport().mapToGlobal(pos))

    def _zoomToError(self, row: int):
        polygonItems = self.drcOutputsModel.getPolygons(row)
        if polygonItems:
            polygonItem = self.drcOutputsModel.getPolygons(row)[0]
            padding = int(getattr(process, 'dbu', 1000)/2)
            self.zoomToRect.emit(polygonItem.polygon().toPolygon().boundingRect().adjusted(-padding, -padding, padding, padding))
            # self.zoomToPolygon.emit(self.drcOutputsModel.getPoints(row))



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#
"""
Type definitions and protocols for editor modules to minimize type errors.
"""

from __future__ import annotations

from typing import Protocol, TYPE_CHECKING, Union

if TYPE_CHECKING:
    import revedaEditor.backend.libBackEnd as libb
    import revedaEditor.backend.libraryModelView as lmview
    from revedaEditor.gui.editorViews import editorView
    from revedaEditor.scenes.editorScene import editorScene

# Forward declarations to avoid circular imports
EditorWindow = Union['schematicEditor', 'layoutEditor', 'symbolEditor']


class EditorContainer(Protocol):
    scene: editorScene  # Base class that all scenes inherit from
    view: editorView  # Base class that all views inherit from
    editorWindow: EditorWindow


class BaseEditor(Protocol):
    """Protocol defining the interface all editors must implement."""

    @property
    def centralW(self) -> EditorContainer: ...

    @property
    def viewItem(self) -> libb.viewItem: ...

    @property
    def libraryDict(self) -> dict: ...

    @property
    def libraryView(self) -> lmview.BaseDesignLibrariesView: ...

    def checkSaveCell(self) -> None: ...

    def saveCell(self) -> None: ...

    def show(self) -> None: ...



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#


from dataclasses import replace

from PySide6.QtCore import (Qt)
from PySide6.QtGui import (QColor, QPen, QBrush)

import revedaEditor.backend.dataDefinitions as ddef

# These layers are builtin layers that are used to create schematic and symbols
# layers are not real layers to draw on but a collection of attributes that a particular
# class of shapes can inherit from.

symbolLayer = ddef.edLayer(name="symbol", pcolor=QColor("green"), pwidth=4, z=2,
                           pstyle=Qt.SolidLine, visible=True, selectable=True)
stretchSymbolLayer = replace(symbolLayer, name="stretchSymbol", pcolor=QColor("red"), z=2)
selectedSymbolLayer = replace(symbolLayer, name="selectedSymbol", pcolor=QColor("blue"),
                              z=3)

symbolPinLayer = ddef.edLayer(name="symbolPin", pcolor=QColor("red"), pwidth=1, z=2,
                              bcolor=QColor("red"), bstyle=Qt.SolidPattern, visible=True,
                              selectable=True)

selectedSymbolPinLayer = replace(symbolPinLayer, name="selectedSymbolPin",
                                 pcolor=QColor("yellow"), z=4)

labelLayer = ddef.edLayer(name="label", pcolor=QColor(255, 255, 153), pwidth=1, z=5,
                          bcolor=QColor(Qt.yellow), bstyle=Qt.SolidPattern,
                          visible=True,
                          selectable=True)
selectedLabelLayer = replace(labelLayer, name="selectedLabel", pcolor=QColor("yellow"), z=6,
                             bcolor=QColor(204, 204, 0))
draftLayer = replace(symbolLayer, pcolor=QColor("gray"), bcolor=QColor("gray"), z=0)

# Symbol Pens
symbolPen = QPen(symbolLayer.pcolor, symbolLayer.pwidth, symbolLayer.pstyle)
symbolPen.setCosmetic(True)
stretchSymbolPen = QPen(stretchSymbolLayer.pcolor, stretchSymbolLayer.pwidth,
                        stretchSymbolLayer.pstyle)
selectedSymbolPen = QPen(selectedSymbolLayer.pcolor, selectedSymbolLayer.pwidth,
                         selectedSymbolLayer.pstyle)

symbolPinPen = QPen(symbolPinLayer.pcolor, symbolPinLayer.pwidth, symbolPinLayer.pstyle)

selectedSymbolPinPen = QPen(selectedSymbolPinLayer.pcolor, selectedSymbolPinLayer.pwidth,
                            selectedSymbolPinLayer.pstyle)

labelPen = QPen(labelLayer.pcolor, labelLayer.pwidth, labelLayer.pstyle)
selectedLabelPen = QPen(selectedLabelLayer.pcolor, selectedLabelLayer.pwidth,
                        selectedLabelLayer.pstyle)
draftPen = QPen(draftLayer.pcolor, draftLayer.pwidth, draftLayer.pstyle)
defaultPen = QPen(draftLayer.pcolor, draftLayer.pwidth, draftLayer.pstyle)

# Brushes
symbolBrush = QBrush(symbolLayer.bcolor, symbolLayer.bstyle)
symbolPinBrush = QBrush(symbolPinLayer.bcolor, symbolPinLayer.bstyle)

selectedSymbolBrush = QBrush(selectedSymbolLayer.bcolor, selectedSymbolLayer.bstyle)

selectedSymbolPinBrush = QBrush(selectedSymbolPinLayer.bcolor,
                                selectedSymbolPinLayer.bstyle)
draftBrush = QBrush(draftLayer.bcolor, draftLayer.bstyle)
labelBrush = QBrush(labelLayer.bcolor, labelLayer.bstyle)
selectedLabelBrush = QBrush(selectedLabelLayer.bcolor, selectedLabelLayer.bstyle)



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#


from typing import List, Tuple, Union

from PySide6.QtCore import QPoint
from PySide6.QtGui import QUndoCommand, QUndoStack
from PySide6.QtWidgets import QGraphicsItem, QGraphicsScene

import revedaEditor.common.layoutShapes as lshp
import revedaEditor.common.shapes as shp


class undoStack(QUndoStack):
    def __init__(self):
        super().__init__()

    def removeLastCommand(self):
        # Remove the last command without undoing it
        if self.canUndo():
            self.setIndex(self.index() - 1)


class addShapeUndo(QUndoCommand):
    def __init__(self, scene: QGraphicsScene, shape: QGraphicsItem):
        super().__init__()
        self._scene = scene
        self._shape = shape
        self.setText("Draw Shape")

    def undo(self):
        self._scene.removeItem(self._shape)
        try:
            self._scene.itemsRefSet.remove(self._shape)
        except KeyError:
            pass

    def redo(self):
        self._scene.addItem(self._shape)
        self._scene.itemsRefSet.add(self._shape)


class addDeleteShapesUndo(QUndoCommand):
    def __init__(
            self,
            scene: QGraphicsScene,
            newShapes: List[QGraphicsItem],
            oldShapes: List[QGraphicsItem],
    ):
        super().__init__()
        self._scene = scene
        self._newShapes = newShapes
        self._oldShapes = oldShapes
        self.setText("Add/Delete Shapes")

    def undo(self):
        for item in self._newShapes:
            self._scene.removeItem(item)
        try:
            self._scene.itemsRefSet -= set(self._newShapes)
        except KeyError:
            pass

        for item in self._oldShapes:
            self._scene.addItem(item)
        self._scene.itemsRefSet |= set(self._oldShapes)

    def redo(self):
        for item in self._newShapes:
            self._scene.addItem(item)
        self._scene.itemsRefSet |= set(self._newShapes)
        for item in self._oldShapes:
            self._scene.removeItem(item)
        try:
            self._scene.itemsRefSet -= set(self._oldShapes)
        except KeyError:
            pass


class addShapesUndo(QUndoCommand):
    def __init__(self, scene: QGraphicsScene, shapes: List[QGraphicsItem]):
        super().__init__()
        self._scene = scene
        self._shapes = shapes
        self.setText("Add Shapes")

    def undo(self):
        for item in self._shapes:
            try:
                self._scene.removeItem(item)
                self._scene.itemsRefSet -= set(self._shapes)
            except Exception as e:
                self._scene.logger.warning(f'Error in undo: {e}')

    def redo(self):
        for item in self._shapes:
            self._scene.addItem(item)
        self._scene.itemsRefSet |= set(self._shapes)


class deleteShapeUndo(QUndoCommand):
    def __init__(self, scene: QGraphicsScene, shape: QGraphicsItem):
        super().__init__()
        self._scene = scene
        self._shape = shape
        self.setText("Delete Shape")

    def undo(self):
        self._scene.addItem(self._shape)

    def redo(self):
        self._scene.removeItem(self._shape)


class deleteShapesUndo(QUndoCommand):
    def __init__(self, scene: QGraphicsScene, shapes: list[QGraphicsItem]):
        super().__init__()
        self._scene = scene
        self._shapes = shapes
        self.setText("Delete Shapes")

    def undo(self):
        [self._scene.addItem(item) for item in self._shapes]

    def redo(self):
        [self._scene.removeItem(item) for item in self._shapes]


class addDeleteShapeUndo(QUndoCommand):
    def __init__(
            self, scene: QGraphicsScene, addShape: QGraphicsItem,
            deleteShape: QGraphicsItem
    ):
        super().__init__()
        self._scene = scene
        self._addshape = addShape
        self._deleteShape = deleteShape
        self.setText("Add/Delete Shape")

    def undo(self):
        self._scene.removeItem(self._addshape)
        self._scene.addItem(self._deleteShape)

    def redo(self):
        self._scene.addItem(self._addshape)
        self._scene.removeItem(self._deleteShape)


class addDeleteNetUndo(QUndoCommand):
    def __init__(
            self, scene: QGraphicsScene, addNet: QGraphicsItem,
            deleteNet: QGraphicsItem
    ):
        super().__init__()
        self._scene = scene
        self._addNet = addNet
        self._deleteNet = deleteNet
        self.setText("Add/Delete Net")

    def undo(self):
        self._scene.removeItem(self._addNet)
        self._scene.addItem(self._deleteNet)

    def redo(self):
        self._scene.addItem(self._addNet)
        self._scene.removeItem(self._deleteNet)


class updateSymUndo(QUndoCommand):
    def __init__(self, item: QGraphicsItem, oldItemList: list, newItemList: list):
        super().__init__()
        self._item = item
        self._oldItemList = oldItemList
        self._newItemList = newItemList
        self.setText("Update Symbol undo")

    def undo(self):
        pass

    def redo(self):
        pass


class moveShapeUndo(QUndoCommand):
    def __init__(
            self,
            scene,
            item: QGraphicsItem,
            attribute: str,
            oldPosition: QPoint,
            newPosition: QPoint,
    ):
        self._scene = scene
        self._item = item
        self._attribute = attribute
        self._oldPosition = oldPosition
        self._newPosition = newPosition
        self.setText("move shape undo")

    def undo(self):
        setattr(self._item, self._attribute, self._oldPosition)

    def redo(self):
        setattr(self._item, self._attribute, self._newPosition)


class undoRotateShape(QUndoCommand):
    def __init__(
            self,
            scene: QGraphicsScene,
            shape: Union[shp.symbolShape, lshp.layoutShape],
            point: QPoint,
            angle: int,
    ):
        super().__init__()
        self._scene = scene
        self._shape = shape
        self._point = point
        self._angle = angle
        self.setText("Undo Shape rotation")

    def undo(self) -> None:
        # self._shape.setRotation(self._angle - 90)
        rotationOriginPoint = self._shape.mapFromScene(self._point)
        self._shape.setTransformOriginPoint(rotationOriginPoint)
        self._shape.angle -= self._angle

    def redo(self) -> None:
        rotationOriginPoint = self._shape.mapFromScene(self._point)
        self._shape.setTransformOriginPoint(rotationOriginPoint)
        self._shape.angle += self._angle


class undoMoveByCommand(QUndoCommand):
    def __init__(
            self, scene, items: List, dx: float, dy: float,
            description: str = "Move Items"
    ):
        super().__init__(description)
        self.scene = scene
        self.items = items
        self.dx = dx
        self.dy = dy
        self.oldPositions: List[Tuple[QGraphicsItem, QPoint]] = []
        self.setText(description)

    def redo(self):
        for item in self.items:
            oldPos = item.pos()
            newPos = oldPos + QPoint(int(self.dx), int(self.dy))
            self.oldPositions.append((item, oldPos))
            item.setPos(newPos)

    def undo(self):
        for item, oldPos in self.oldPositions:
            item.setPos(oldPos)


class undoGroupMove(QUndoCommand):
    def __init__(self, scene, items: List, oldPosList: List[QPoint],
                 posDiff: QPoint):
        super().__init__()
        self.scene = scene
        self.items = items
        self.oldPosList = oldPosList
        self.posDiff = posDiff
        self.setText("Undo Group Move")

    def redo(self):
        for item, oldPos in zip(self.items, self.oldPosList):
            item.setPos(oldPos + self.posDiff)

    def undo(self):
        for item, oldPos in zip(self.items, self.oldPosList):
            item.setPos(oldPos)



""" Console
Interactive console widget.  Use to add an interactive python interpreter
in a GUI application.
"""

#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

import code
import re
from typing import Callable

from PySide6.QtCore import (
    Qt,
    Signal,
    QEvent,
    QSize,
)
from PySide6.QtGui import QTextCharFormat, QBrush, QColor, QFont
from PySide6.QtWidgets import QLineEdit, QWidget, QGridLayout, QPlainTextEdit


class LineEdit(QLineEdit):
    """QLIneEdit with a history buffer for recalling previous lines.
    I also accept tab as input (4 spaces).
    """

    newline = Signal(str)  # Signal when return key pressed

    def __init__(self, history: int = 100) -> "LineEdit":
        super().__init__()
        self.historymax = history
        self.clearhistory()
        self.promptpattern = re.compile("^[>.]")

    def clearhistory(self) -> None:
        """Clear history buffer"""
        self.historyindex = 0
        self.historylist = []

    def event(self, ev: QEvent) -> bool:
        """Intercept tab and arrow key presses.  Insert 4 spaces
        when tab pressed instead of moving to next contorl.  WHen
        arrow up or down are pressed select a line from the history
        buffer.  Emit newline signal when return key is pressed.
        """
        if ev.type() == QEvent.KeyPress:
            if ev.key() == int(Qt.Key.Key_Tab):
                self.insert("    ")
                return True
            elif ev.key() == int(Qt.Key.Key_Up):
                self.recall(self.historyindex - 1)
                return True
            elif ev.key() == int(Qt.Key.Key_Down):
                self.recall(self.historyindex + 1)
                return True
            elif ev.key() == int(Qt.Key.Key_Home):
                self.recall(0)
                return True
            elif ev.key() == int(Qt.Key.Key_End):
                self.recall(len(self.historylist) - 1)
                return True
            elif ev.key() == int(Qt.Key.Key_Return):
                self.returnkey()
                return True
        return super().event(ev)

    def returnkey(self) -> None:
        """Return key was pressed.  Add line to history and emit
        the newline signal.
        """
        text = self.text().rstrip()
        self.record(text)
        self.newline.emit(text)
        self.setText("")

    def recall(self, index: int) -> None:
        """Select a line from the history list"""
        length = len(self.historylist)
        if length > 0:
            index = max(0, min(index, length - 1))
            self.setText(self.historylist[index])
            self.historyindex = index

    def record(self, line: str) -> None:
        """Add line to history buffer"""
        self.historyindex += 1
        while len(self.historylist) >= self.historymax - 1:
            self.historylist.pop()
        self.historylist.append(line)
        self.historyindex = min(self.historyindex, len(self.historylist))


class Redirect:
    """Map self.write to a function"""

    def __init__(self, func: Callable) -> "Redirect":
        self.func = func

    def write(self, line: str) -> None:
        self.func(line)


class pythonConsole(QWidget):
    """A GUI version of code.InteractiveConsole."""

    def __init__(
            self,
            context=locals(),  # context for interpreter
            history: int = 200,  # max lines in history buffer
            blockcount: int = 500,  # max lines in output buffer
    ) -> "pythonConsole":
        super().__init__()
        self.setcontext(context)
        self.buffer = []

        self.content = QGridLayout(self)
        self.content.setContentsMargins(0, 0, 0, 0)
        self.content.setSpacing(0)

        # Display for output and stderr
        self.outdisplay = QPlainTextEdit(self)
        self.outdisplay.setMaximumBlockCount(blockcount)
        self.outdisplay.setReadOnly(True)
        self.content.addWidget(self.outdisplay, 0, 0, 1, 2)

        # Use color to differentiate input, output and stderr
        self.inpfmt = self.outdisplay.currentCharFormat()
        self.outfmt = QTextCharFormat(self.inpfmt)
        self.outfmt.setForeground(QBrush(QColor(0, 0, 255)))
        self.errfmt = QTextCharFormat(self.inpfmt)
        self.errfmt.setForeground(QBrush(QColor(255, 0, 0)))

        # Display input prompt left of input edit
        self.promptdisp = QLineEdit(self)
        self.promptdisp.setReadOnly(True)
        self.promptdisp.setFixedWidth(15)
        self.promptdisp.setFrame(False)
        self.content.addWidget(self.promptdisp, 1, 0)
        self.setprompt("> ")

        # Enter commands here
        self.inpedit = LineEdit(history=history)
        self.inpedit.newline.connect(self.push)
        self.inpedit.setFrame(False)
        self.content.addWidget(self.inpedit, 1, 1)

    def setcontext(self, context):
        """Set context for interpreter"""
        self.interp = code.InteractiveInterpreter(context)

    def resetbuffer(self) -> None:
        """Reset the input buffer."""
        self.buffer = []

    def setprompt(self, text: str):
        self.prompt = text
        self.promptdisp.setText(text)

    def push(self, line: str) -> None:
        """Execute entered command.  Command may span multiple lines"""
        if line == "clear":
            self.inpedit.clearhistory()
            self.outdisplay.clear()
        else:
            lines = line.split("\n")
            for line in lines:
                if re.match("^[>.] ", line):
                    line = line[2:]
                self.writeoutput(self.prompt + line, self.inpfmt)
                self.setprompt(". ")
                self.buffer.append(line)
            # Built a command string from lines in the buffer
            source = "\n".join(self.buffer)
            more = self.interp.runsource(source, "<input>", "exec")
            if not more:
                self.setprompt("> ")
                self.resetbuffer()

    def setfont(self, font: QFont) -> None:
        """Set font for input and display widgets.  Should be monospaced"""
        self.outdisplay.setFont(font)
        self.inpedit.setFont(font)

    def write(self, line: str) -> None:
        """Capture stdout and display in outdisplay"""
        if len(line) != 1 or ord(line[0]) != 10:
            self.writeoutput(line.rstrip(), self.outfmt)

    def errorwrite(self, line: str) -> None:
        """Capture stderr and display in outdisplay"""
        self.writeoutput(line, self.errfmt)

    def writeoutput(self, line: str, fmt: QTextCharFormat = None) -> None:
        """Set text formatting and display line in outdisplay"""
        if fmt is not None:
            self.outdisplay.setCurrentCharFormat(fmt)
        self.outdisplay.appendPlainText(line.rstrip())

    def sizeHint(self):
        return QSize(500, 200)



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause Lic\ense Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#


import datetime
import pathlib

from PySide6.QtCore import Qt, QDir
from PySide6.QtGui import QStandardItemModel, QStandardItem
from PySide6.QtWidgets import (
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFileDialog,
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QVBoxLayout,
    QRadioButton,
    QButtonGroup,
    QPushButton,
    QGroupBox,
    QTableView,
    QMenu,
    QCheckBox,
)

import revedaEditor.backend.libraryMethods as libm
import revedaEditor.gui.editFunctions as edf


class createCellDialog(QDialog):
    def __init__(self, parent, model):
        super().__init__(parent=parent)
        self.parent = parent
        self.model = model
        self.init_UI()

    def init_UI(self):
        self.setWindowTitle("Create Cell")
        self.layout = QFormLayout()
        self.layout.setSpacing(10)
        self.libNamesCB = QComboBox()
        self.libNamesCB.setModel(self.model)
        self.libNamesCB.setModelColumn(0)
        self.libNamesCB.setCurrentIndex(0)
        self.libNamesCB.currentTextChanged.connect(self.selectLibrary)
        self.layout.addRow(edf.boldLabel("Library:"), self.libNamesCB)
        self.cellCB = QComboBox()
        libItem = libm.getLibItem(self.model, self.libNamesCB.currentText())
        self.cellList = sorted(
            [libItem.child(i).cellName for i in range(libItem.rowCount())])
        self.cellCB.addItems(self.cellList)
        self.cellCB.setEditable(True)
        self.layout.addRow(edf.boldLabel("Cell Name:"), self.cellCB)
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

    def selectLibrary(self):
        libItem = libm.getLibItem(self.model, self.libNamesCB.currentText())
        cellList = sorted([libItem.child(i).cellName for i in range(libItem.rowCount())])
        self.cellCB.clear()
        self.cellCB.addItems(cellList)


class deleteCellDialog(createCellDialog):
    def __init__(self, parent, model):
        super().__init__(parent, model)
        self.cellCB.setEditable(False)
        self.setWindowTitle("Delete Cell")


class newCellViewDialog(createCellDialog):
    def __init__(self, parent, model):
        super().__init__(parent, model)
        self.cellCB.setEditable(False)
        self.setWindowTitle("Create Cell View")
        self.viewType = QComboBox()
        self.viewType.currentIndexChanged.connect(self.setCurrentViewName)
        self.layout.addRow(edf.boldLabel("View Type:"), self.viewType)
        self.viewName = edf.longLineEdit()
        self.layout.addRow(edf.boldLabel("View Name:"), self.viewName)
        self.layout.setSpacing(10)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

    def setCurrentViewName(self):
        self.viewName.setText(self.viewType.currentText())


class selectCellViewDialog(createCellDialog):
    def __init__(self, parent, model):
        super().__init__(parent=parent, model=model)
        libItem = libm.getLibItem(self.model, self.libNamesCB.currentText())
        self.setWindowTitle("Select CellView")
        self.cellCB.setEditable(False)
        self.cellCB.currentTextChanged.connect(self.cellNameChanged)
        self.viewCB = QComboBox()
        cellItem = libm.getCellItem(libItem, self.cellCB.currentText())
        # self.viewCB.addItems(
        #     [cellItem.child(i).text() for i in range(cellItem.rowCount())]
        # )
        self.viewCB.addItems(
            sorted([cellItem.child(i).text() for i in range(cellItem.rowCount())])
        )
        self.layout.addRow(edf.boldLabel("View Name:"), self.viewCB)
        self.layout.setSpacing(10)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

    def cellNameChanged(self):
        libItem = libm.getLibItem(self.model, self.libNamesCB.currentText())
        cellItem = libm.getCellItem(libItem, self.cellCB.currentText())
        if cellItem is not None:
            viewList = [cellItem.child(i).text() for i in range(cellItem.rowCount())]
        else:
            viewList = []
        self.viewCB.clear()
        self.viewCB.addItems(viewList)


class renameCellDialog(QDialog):
    def __init__(self, parent, cellItem):
        super().__init__(parent=parent)
        self.parent = parent
        self.cellItem = cellItem

        self.init_UI()

    def init_UI(self):
        self.setWindowTitle("Rename Cell")
        layout = QFormLayout()
        layout.setSpacing(10)
        self.nameEdit = QLineEdit()
        self.nameEdit.setPlaceholderText("Cell Name")
        self.nameEdit.setFixedWidth(200)
        layout.addRow(edf.boldLabel("Cell Name:"), self.nameEdit)
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        layout.addRow(self.buttonBox)
        self.setLayout(layout)


class copyCellDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent=parent)
        self.parent = parent

        # self.index = 0
        self.init_UI()

    def init_UI(self):
        self.setWindowTitle("Copy Cell")
        layout = QFormLayout()
        layout.setSpacing(10)
        self.libraryCB = QComboBox()
        self.selectedLibPath = self.libraryCB.itemData(0, Qt.ItemDataRole.UserRole + 2)
        self.libraryCB.currentTextChanged.connect(self.selectLibrary)
        layout.addRow(edf.boldLabel("Library:"), self.libraryCB)
        self.copyName = QLineEdit()
        self.copyName.setPlaceholderText("Enter Cell Name")
        self.copyName.setFixedWidth(130)
        layout.addRow(edf.boldLabel("Cell Name:"), self.copyName)
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        layout.addRow(self.buttonBox)
        self.setLayout(layout)

    def selectLibrary(self):
        self.selectedLibPath = self.libraryCB.itemData(
            self.libraryCB.currentIndex(), Qt.ItemDataRole.UserRole + 2
        )


class copyViewDialog(createCellDialog):
    def __init__(self, parent, model):
        super().__init__(parent=parent, model=model)
        self.setWindowTitle("Copy View")
        self.cellCB.setEditable(True)
        self.cellCB.InsertPolicy = QComboBox.InsertAfterCurrent
        self.viewName = edf.longLineEdit()
        self.layout.addRow(edf.boldLabel("View Name:"), self.viewName)
        self.layout.setSpacing(10)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)


class closeLibDialog(QDialog):
    def __init__(self, libraryDict, parent, *args):
        super().__init__(parent, *args)
        self.libraryDict = libraryDict
        self.setWindowTitle("Select Library to close")
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        layout = QVBoxLayout()
        formLayout = QFormLayout()
        self.libNamesCB = QComboBox()
        self.libNamesCB.addItems(self.libraryDict.keys())
        formLayout.addRow(edf.boldLabel("Select Library", self), self.libNamesCB)
        layout.addLayout(formLayout)
        layout.addSpacing(40)
        layout.addWidget(self.buttonBox)
        self.setLayout(layout)


class renameLibDialog(QDialog):
    def __init__(self, parent, oldLibraryName, *args):
        super().__init__(parent, *args)
        self.oldLibraryName = oldLibraryName
        self.setWindowTitle(f"Change {oldLibraryName} to:")
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)

        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        formBox = QGroupBox("Rename Library")
        layout = QVBoxLayout()
        formLayout = QFormLayout()
        self.newLibraryName = edf.longLineEdit()
        formLayout.addRow(edf.boldLabel("New Library Name:", self), self.newLibraryName)
        formBox.setLayout(formLayout)
        layout.addWidget(formBox)
        layout.addSpacing(40)
        layout.addWidget(self.buttonBox)
        self.setLayout(layout)


class renameViewDialog(QDialog):
    def __init__(self, parent, oldViewName):
        super().__init__(parent)
        self.oldViewName = oldViewName
        self.setWindowTitle(f"Rename {oldViewName} ")
        self.layout = QVBoxLayout()
        formLayout = QFormLayout()
        oldViewNameEdit = edf.longLineEdit()
        oldViewNameEdit.setText(self.oldViewName)
        oldViewNameEdit.setEnabled(False)
        formLayout.addRow(edf.boldLabel("Old View Name:"), oldViewNameEdit)
        self.newViewNameEdit = edf.longLineEdit()
        formLayout.addRow(edf.boldLabel("New View Name:"), self.newViewNameEdit)
        self.layout.addLayout(formLayout)
        self.layout.setSpacing(10)
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)


class deleteSymbolDialog(QDialog):
    def __init__(self, cellName, viewName, *args):
        super().__init__(*args)
        self.setWindowTitle(f"Delete {cellName}-{viewName} CellView?")
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.layout = QVBoxLayout()
        message = QLabel(f"{cellName}-{viewName} will be recreated!")
        self.layout.addWidget(message)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)


class netlistExportDialogue(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setWindowTitle(f"Export Netlist for {parent.cellName}-{parent.viewName}")
        # self.setMinimumSize(500, 100)
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.mainLayout = QVBoxLayout()
        self.mainLayout.addStretch(2)
        viewBox = QGroupBox("Select a view to Netlist")
        viewBoxLayout = QFormLayout()
        self.libNameEdit = edf.longLineEdit()
        self.libNameEdit.setDisabled(True)
        viewBoxLayout.addRow(edf.boldLabel("Library:"), self.libNameEdit)
        self.cellNameEdit = edf.longLineEdit()
        self.cellNameEdit.setDisabled(True)
        viewBoxLayout.addRow(edf.boldLabel("Cell:"), self.cellNameEdit)
        self.viewNameCombo = QComboBox()
        viewBoxLayout.addRow(edf.boldLabel("View:"), self.viewNameCombo)
        viewBox.setLayout(viewBoxLayout)
        self.mainLayout.addWidget(viewBox)
        switchBox = QGroupBox("Switch and Stop View Lists")
        self.formLayout = QFormLayout()
        self.switchViewEdit = edf.longLineEdit()
        self.switchViewEdit.setText((", ").join(self.parent.switchViewList))
        self.formLayout.addRow(edf.boldLabel("Switch View List:"), self.switchViewEdit)
        self.stopViewEdit = edf.longLineEdit()
        self.stopViewEdit.setText((", ").join(self.parent.stopViewList))
        self.formLayout.addRow((edf.boldLabel("Stop View: ")), self.stopViewEdit)
        switchBox.setLayout(self.formLayout)
        self.mainLayout.addWidget(switchBox)
        netlistOptionBox = QGroupBox("Netlist Options")
        netlistOptLayout = QVBoxLayout()
        self.topAsSubcktCheckBox = QCheckBox("Netlist top level as subcircuit")
        self.topAsSubcktCheckBox.setChecked(False)
        netlistOptLayout.addWidget(self.topAsSubcktCheckBox)
        netlistOptionBox.setLayout(netlistOptLayout)
        self.mainLayout.addWidget(netlistOptionBox)
        fileBox = QGroupBox("Select Simulation Directory")
        fileDialogLayout = QHBoxLayout()
        fileDialogLayout.addWidget(edf.boldLabel("Export Directory:"))
        self.netlistDirEdit = edf.longLineEdit()
        fileDialogLayout.addWidget(self.netlistDirEdit)
        self.netListDirButton = QPushButton("...")
        self.netListDirButton.clicked.connect(self.onDirButtonClicked)
        fileDialogLayout.addWidget(self.netListDirButton)
        fileBox.setLayout(fileDialogLayout)
        self.mainLayout.addWidget(fileBox)
        self.mainLayout.addStretch(2)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)

    def onDirButtonClicked(self):
        self.dirName = QFileDialog.getExistingDirectory()
        if self.dirName:
            self.netlistDirEdit.setText(self.dirName)


class layoutExportDialogue(QDialog):
    def __init__(self, parentW, export_format="GDS"):
        super().__init__(parentW)
        self.parentW = parentW
        self.export_format = export_format.upper()
        self.setMinimumWidth(500)
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.mainLayout = QVBoxLayout()
        self.mainLayout.addStretch(2)
        settingsBox = QGroupBox(f"{self.export_format} Export Settings")
        settingsBoxLayout = QFormLayout()
        settingsBox.setLayout(settingsBoxLayout)
        self.unitEdit = edf.shortLineEdit()
        self.unitEdit.setToolTip(f"The unit of the {self.export_format} file.")
        settingsBoxLayout.addRow(edf.boldLabel("Unit:"), self.unitEdit)
        self.precisionEdit = edf.shortLineEdit()
        self.precisionEdit.setToolTip(f"The precision of the {self.export_format} file.")
        settingsBoxLayout.addRow(edf.boldLabel("Precision:"), self.precisionEdit)
        self.mainLayout.addWidget(settingsBox)
        fileBox = QGroupBox(f"{self.export_format} Export Directory")
        fileDialogLayout = QHBoxLayout()
        fileDialogLayout.addWidget(edf.boldLabel("Export Directory/File:"))
        self.exportPathEdit = edf.longLineEdit()
        fileDialogLayout.addWidget(self.exportPathEdit)
        self.exportButton = QPushButton("...")
        self.exportButton.clicked.connect(self.onDirButtonClicked)
        fileDialogLayout.addWidget(self.exportButton)
        fileBox.setLayout(fileDialogLayout)
        self.mainLayout.addWidget(fileBox)
        self.mainLayout.addStretch(2)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)

    def onDirButtonClicked(self):
        dirName = QFileDialog.getExistingDirectory()
        if dirName:
            self.exportPathEdit.setText(
                f"{dirName}/{self.parentW.cellName}"
            )


class gdsExportDialogue(layoutExportDialogue):
    def __init__(self, parentW):
        super().__init__(parentW, "GDS")


class oasExportDialogue(layoutExportDialogue):
    def __init__(self, parentW):
        super().__init__(parentW, "OAS")

class gdsImportDialogue(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setWindowTitle(f"Import GDS File")
        self.setMinimumWidth(500)
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.mainLayout = QVBoxLayout()
        self.mainLayout.addStretch(2)
        settingsBox = QGroupBox("GDS Import Settings")
        settingsBoxLayout = QFormLayout()
        settingsBox.setLayout(settingsBoxLayout)
        self.libNameEdit = edf.longLineEdit()
        self.libNameEdit.setToolTip("The name of the library to import the GDS into.")
        settingsBoxLayout.addRow(edf.boldLabel("Import Library Name:"), self.libNameEdit)
        self.unitEdit = edf.shortLineEdit()
        self.unitEdit.setToolTip("The unit of the GDS file.")
        settingsBoxLayout.addRow(edf.boldLabel("Unit:"), self.unitEdit)
        self.precisionEdit = edf.shortLineEdit()
        self.precisionEdit.setToolTip("The precision of the GDS file.")
        settingsBoxLayout.addRow(edf.boldLabel("Precision:"), self.precisionEdit)
        self.mainLayout.addWidget(settingsBox)
        fileBox = QGroupBox("GDS File")
        fileDialogLayout = QHBoxLayout()
        fileDialogLayout.addWidget(edf.boldLabel("GDS File:"))
        self.inputFileEdit = edf.longLineEdit()
        fileDialogLayout.addWidget(self.inputFileEdit)
        self.gdsImportButton = QPushButton("...")
        self.gdsImportButton.clicked.connect(self.onFileButtonClicked)
        fileDialogLayout.addWidget(self.gdsImportButton)
        fileBox.setLayout(fileDialogLayout)
        self.mainLayout.addWidget(fileBox)
        self.mainLayout.addStretch(2)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)

    def onFileButtonClicked(self):
        gdsFileName, _ = QFileDialog.getOpenFileName(self, caption="Select GDS file.",
                                                     filter="GDS files (*.gds)")
        if gdsFileName:
            self.inputFileEdit.setText(
                gdsFileName
            )


class goDownHierDialogue(QDialog):
    def __init__(
            self,
            parent,
    ):
        super().__init__(parent=parent)
        self._parent = parent
        self.setWindowTitle("Go Down Hierarchy")
        self.setMinimumWidth(250)
        self.buttonId = 1
        _mainLayout = QVBoxLayout()
        viewGroup = QGroupBox("Select a cellview")
        viewGroupLayout = QVBoxLayout()
        viewGroup.setLayout(viewGroupLayout)
        self.viewListCB = QComboBox()
        viewGroupLayout.addWidget(self.viewListCB)
        _mainLayout.addWidget(viewGroup)
        buttonGroupBox = QGroupBox("Open")
        buttonGroupLayout = QHBoxLayout()
        buttonGroupBox.setLayout(buttonGroupLayout)
        self.openButton = QRadioButton("Edit")
        self.openButton.setChecked(True)
        self.readOnlyButton = QRadioButton("Read Only")
        buttonGroupLayout.addWidget(self.openButton)
        buttonGroupLayout.addWidget(self.readOnlyButton)
        _mainLayout.addWidget(buttonGroupBox)
        self.buttonGroup = QButtonGroup()
        self.buttonGroup.addButton(self.openButton, id=1)
        self.buttonGroup.addButton(self.readOnlyButton, id=2)
        self.buttonGroup.buttonClicked.connect(self.onButtonClicked)
        _mainLayout.addWidget(buttonGroupBox)
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        buttonBox = QDialogButtonBox(QBtn)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)
        _mainLayout.addWidget(buttonBox)
        self.setLayout(_mainLayout)

    def onButtonClicked(self):
        self.buttonId = self.buttonGroup.checkedId()


class importCellDialogue(QDialog):
    def __init__(self, model, parent, file_type="Verilog-A"):
        super().__init__(parent)
        self._parent = parent
        self.file_type = file_type
        self._model = model

        # Configure dialog based on file type
        if file_type == "Verilog-A":
            self.setWindowTitle("Import a Verilog-a Module File")
            self.setMinimumSize(500, 400)
            self.file_extension = ".va"
            self.file_filter = "Verilog-A files (*.va)"
            self.file_label = "Select Verilog-A file:"
            self.view_label = "Verilog-A view:"
            self.caption = "Select Verilog-A file."
        elif file_type == "Spice":
            self.setWindowTitle("Import a Spice Subcircuit File")
            self.setMinimumSize(500, 200)
            self.file_extension = ".sp"
            self.file_filter = "Spice files (*.sp, *.cir)"
            self.file_label = "Select Spice file:"
            self.view_label = "Spice  cellview:"
            self.caption = "Select Spice file."

        self._setup_ui()
        self.show()

    def _setup_ui(self):
        mainLayout = QVBoxLayout()
        fileDialogLayout = QHBoxLayout()
        fileDialogLayout.addWidget(edf.boldLabel(self.file_label), 1)
        self.fileEdit = edf.longLineEdit()
        fileDialogLayout.addWidget(self.fileEdit, 4)
        self.fileButton = QPushButton("...")
        self.fileButton.clicked.connect(self.onFileButtonClicked)
        fileDialogLayout.addWidget(self.fileButton, 1)
        mainLayout.addLayout(fileDialogLayout)
        mainLayout.addSpacing(20)

        layout = QFormLayout()
        layout.setSpacing(10)
        self.libNamesCB = QComboBox()
        self.libNamesCB.setModel(self._model)
        self.libNamesCB.currentTextChanged.connect(self.changeCells)
        layout.addRow(edf.boldLabel("Library:"), self.libNamesCB)
        self.cellNamesCB = QComboBox()
        self.cellNamesCB.setEditable(True)

        # Handle initial cell names
        try:
            if self.file_type == "Verilog-A":
                try:
                    initialCellNames = [
                        self._model.item(0).child(i).cellName
                        for i in range(self._model.item(0).rowCount())
                    ]
                except Exception as e:
                    initialCellNames = []
                    print(f'No libraries could be found.')
            else:  # Spice
                initialCellNames = [
                    self._model.item(0).child(i).cellName
                    for i in range(self._model.item(0).rowCount())
                ]
        except (AttributeError, IndexError):
            initialCellNames = []

        self.cellNamesCB.addItems(initialCellNames)
        layout.addRow(edf.boldLabel("Cell:"), self.cellNamesCB)
        self.viewName = edf.longLineEdit()
        layout.addRow(edf.boldLabel(self.view_label), self.viewName)
        mainLayout.addLayout(layout)

        symbolGroupBox = QGroupBox("Symbol Creation")
        symbolGBLayout = QVBoxLayout()
        self.symbolCheckBox = QCheckBox("Create a new symbol?")
        symbolGBLayout.addWidget(self.symbolCheckBox)
        symbolGroupBox.setLayout(symbolGBLayout)
        mainLayout.addWidget(symbolGroupBox)
        mainLayout.addSpacing(20)

        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        mainLayout.addWidget(self.buttonBox)
        self.setLayout(mainLayout)

    def changeCells(self):
        selectedLibItemRow = self._model.findItems(self.libNamesCB.currentText())[
            0
        ].row()
        libCellNames = [
            self._model.item(selectedLibItemRow).child(i).cellName
            for i in range(self._model.item(selectedLibItemRow).rowCount())
        ]
        self.cellNamesCB.clear()
        self.cellNamesCB.addItems(libCellNames)

    def onFileButtonClicked(self):
        filePathObj = pathlib.Path(self.fileEdit.text())
        fileDialog = QFileDialog()
        fileDialog.setNameFilter(self.file_filter)
        fileDialog.setDirectory(str(filePathObj.parent))
        fileDialog.selectFile(filePathObj.name)
        fileName = fileDialog.getOpenFileName(
            self, caption=self.caption
        )[0]
        if fileName:
            self.fileEdit.setText(fileName)


class importVerilogaCellDialogue(importCellDialogue):
    def __init__(self, model, parent):
        super().__init__(model, parent, "Verilog-A")

    @property
    def vaFileEdit(self):
        """Compatibility property for existing code"""
        return self.fileEdit

    @property
    def vaViewName(self):
        """Compatibility property for existing code"""
        return self.viewName

    @property
    def vaFileName(self):
        """Compatibility property for existing code"""
        return getattr(self, '_fileName', None)

    @vaFileName.setter
    def vaFileName(self, value):
        self._fileName = value


class importSpiceCellDialogue(importCellDialogue):
    def __init__(self, model, parent):
        super().__init__(model, parent, "Spice")

    @property
    def spiceFileEdit(self):
        """Compatibility property for existing code"""
        return self.fileEdit

    @property
    def spiceViewName(self):
        """Compatibility property for existing code"""
        return self.viewName

    @property
    def spiceFileName(self):
        """Compatibility property for existing code"""
        return getattr(self, '_fileName', None)

    @spiceFileName.setter
    def spiceFileName(self, value):
        self._fileName = value


class createConfigViewDialogue(QDialog):
    def __init__(self, parent):
        super().__init__(parent=parent)
        self.parent = parent
        self.mainLayout = QVBoxLayout()
        self.setWindowTitle("Create New Config View")
        self.setMinimumSize(360, 400)
        topCellGroup = QGroupBox("Top Cell")
        topCellLayout = QFormLayout()
        self.libraryNameEdit = edf.longLineEdit()
        topCellLayout.addRow(edf.boldLabel("Library:"), self.libraryNameEdit)
        self.cellNameEdit = edf.longLineEdit()
        topCellLayout.addRow(edf.boldLabel("Cell:"), self.cellNameEdit)
        self.viewNameCB = QComboBox()
        topCellLayout.addRow(edf.boldLabel("View:"), self.viewNameCB)
        topCellGroup.setLayout(topCellLayout)
        self.mainLayout.addWidget(topCellGroup)
        viewGroup = QGroupBox("Switch/Stop Views")
        viewGroupLayout = QFormLayout()
        viewGroup.setLayout(viewGroupLayout)
        self.switchViews = edf.longLineEdit()
        viewGroupLayout.addRow(edf.boldLabel("View List:"), self.switchViews)
        self.stopViews = edf.longLineEdit()
        viewGroupLayout.addRow(edf.boldLabel("Stop List:"), self.stopViews)
        self.mainLayout.addWidget(viewGroup)
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)


class appProperties(QDialog):
    def __init__(self, parent):
        self.parent = parent
        super().__init__(parent)
        self.setMinimumSize(750, 550)
        self.setWindowTitle("Revolution EDA Options")
        mainLayout = QVBoxLayout()
        mainLayout.setStretch(0, 2)
        mainLayout.setStretch(1, 2)
        filePathsGroup = QGroupBox("Paths")
        filePathsLayout = QVBoxLayout()
        filePathsLayout.setSpacing(20)

        rootPathDialogLayout = QHBoxLayout()
        rootPathDialogLayout.addWidget(edf.boldLabel("Root (Run) Path:"), 2)
        self.rootPathEdit = edf.longLineEdit()
        rootPathDialogLayout.addWidget(self.rootPathEdit, 5)
        self.rootPathButton = QPushButton("...")
        self.rootPathButton.clicked.connect(self.onRootPathButtonClicked)
        filePathsLayout.addLayout(rootPathDialogLayout)
        rootPathDialogLayout.addWidget(self.rootPathButton, 1)
        simInPathDialogLayout = QHBoxLayout()
        simInPathDialogLayout.addWidget(edf.boldLabel("PDK Path:"), 2)
        self.simInpPathEdit = edf.longLineEdit()
        simInPathDialogLayout.addWidget(self.simInpPathEdit, 5)
        self.simInpPathButton = QPushButton("...")
        self.simInpPathButton.clicked.connect(self.onSimInpPathButtonClicked)
        simInPathDialogLayout.addWidget(self.simInpPathButton, 1)
        filePathsLayout.addLayout(simInPathDialogLayout)
        simOutPathDialogLayout = QHBoxLayout()
        simOutPathDialogLayout.addWidget(edf.boldLabel("Simulation Outputs Path:"), 2)
        self.simOutPathEdit = edf.longLineEdit()
        simOutPathDialogLayout.addWidget(self.simOutPathEdit, 5)
        self.simOutPathButton = QPushButton("...")
        self.simOutPathButton.clicked.connect(self.onSimOutPathButtonClicked)
        simOutPathDialogLayout.addWidget(self.simOutPathButton, 1)
        filePathsLayout.addLayout(simOutPathDialogLayout)
        pluginsPathDialogLayout = QHBoxLayout()
        pluginsPathDialogLayout.addWidget(edf.boldLabel("Plugins Path:"), 2)
        self.pluginsPathEdit = edf.longLineEdit()
        pluginsPathDialogLayout.addWidget(self.pluginsPathEdit, 5)
        self.pluginsPathButton = QPushButton("...")
        self.pluginsPathButton.clicked.connect(self.onPluginsPathButtonClicked)
        pluginsPathDialogLayout.addWidget(self.pluginsPathButton, 1)
        filePathsLayout.addLayout(pluginsPathDialogLayout)
        vaModulePathDialogLayout = QHBoxLayout()
        vaModulePathDialogLayout.addWidget(edf.boldLabel("Verilog-A Module Path:"), 2)
        self.vaModulePathEdit = edf.longLineEdit()
        vaModulePathDialogLayout.addWidget(self.vaModulePathEdit, 5)
        self.vaModulePathButton = QPushButton("...")
        self.vaModulePathButton.clicked.connect(self.onVaModulePathButtonClicked)
        vaModulePathDialogLayout.addWidget(self.vaModulePathButton, 1)
        filePathsLayout.addLayout(vaModulePathDialogLayout)
        filePathsGroup.setLayout(filePathsLayout)
        mainLayout.addWidget(filePathsGroup)
        switchViewsGroup = QGroupBox("Switch and Stop Views")
        switchViewsLayout = QFormLayout()
        switchViewsLayout.setSpacing(20)
        self.switchViewsEdit = edf.longLineEdit()
        switchViewsLayout.addRow(edf.boldLabel("Switch Views:"), self.switchViewsEdit)
        self.stopViewsEdit = edf.longLineEdit()
        switchViewsLayout.addRow(edf.boldLabel("Stop Views:"), self.stopViewsEdit)
        switchViewsGroup.setLayout(switchViewsLayout)
        mainLayout.addWidget(switchViewsGroup)
        performanceGroup = QGroupBox("Performance Settings")
        performanceLayout = QFormLayout()
        self.threadPoolEdit = edf.shortLineEdit()
        performanceLayout.addRow(edf.boldLabel("Thread Pool Max Count:"),
                                 self.threadPoolEdit)
        performanceGroup.setLayout(performanceLayout)
        mainLayout.addWidget(performanceGroup)
        saveGroupBox = QGroupBox("Save Options")
        saveGBLayout = QVBoxLayout()
        self.optionSaveBox = QCheckBox("Save options to configuration file?")
        saveGBLayout.addWidget(self.optionSaveBox)
        saveGroupBox.setLayout(saveGBLayout)
        mainLayout.addWidget(saveGroupBox)
        mainLayout.addSpacing(20)
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        mainLayout.addWidget(self.buttonBox)
        self.setLayout(mainLayout)

    def onRootPathButtonClicked(self):
        self.rootPathEdit.setText(
            QFileDialog.getExistingDirectory(self, caption="Root run path:")
        )

    def onSimOutPathButtonClicked(self):
        self.simOutPathEdit.setText(
            QFileDialog.getExistingDirectory(self, caption="Simulation Outputs path:")
        )

    def onSimInpPathButtonClicked(self):
        self.simInpPathEdit.setText(
            QFileDialog.getExistingDirectory(self, caption="Simulation Inputs (PDK) path:")
        )

    def onPluginsPathButtonClicked(self):
        self.pluginsPathEdit.setText(
            QFileDialog.getExistingDirectory(self, caption="Plugins path:")
        )

    def onVaModulePathButtonClicked(self):
        self.vaModulePathEdit.setText(
            QFileDialog.getExistingDirectory(self, caption="Verilog-a Modules path:")
        )


class libraryPathsModel(QStandardItemModel):
    def __init__(self, libraryDict):
        super().__init__()
        self.libraryDict = libraryDict
        self.setHorizontalHeaderLabels(["Library Name", "Library Path"])
        for key, value in self.libraryDict.items():
            libName = QStandardItem(key)
            libPath = QStandardItem(str(value))
            self.appendRow([libName, libPath])
        self.appendRow([QStandardItem("Right click here..."), QStandardItem("")])


class libraryPathsTableView(QTableView):
    def __init__(self, model, logger):
        super().__init__()
        self.model = model
        self.logger = logger
        self.setModel(self.model)
        self.setShowGrid(True)
        self.setColumnWidth(0, 200)
        self.setColumnWidth(1, 400)
        self.fileDialog = QFileDialog()
        self.fileDialog.setFileMode(QFileDialog.Directory)
        self.libNameEditList = list()
        self.libPathEditList = list()
        for row in range(self.model.rowCount()):
            self.libPathEditList.append(edf.longLineEdit())
            self.setIndexWidget(self.model.index(row, 1), self.libPathEditList[-1])
            self.libPathEditList[-1].setText(self.model.item(row, 1).text())

    def contextMenuEvent(self, event) -> None:
        self.menu = QMenu(self)
        try:
            selectedIndex = self.selectedIndexes()[0]
        except IndexError:
            self.model.appendRow([QStandardItem("Click here..."), QStandardItem("")])
            selectedIndex = self.model.index(0, 0)
        removePathAction = self.menu.addAction("Remove Path")
        removePathAction.triggered.connect(
            lambda: self.removeLibraryPath(selectedIndex)
        )
        addPathAction = self.menu.addAction("Add Library Path")
        addPathAction.triggered.connect(lambda: self.addLibraryPath(selectedIndex))
        self.menu.exec(event.globalPos())

    def removeLibraryPath(self, index):
        self.model.takeRow(index.row())
        self.logger.info("Removed Library Path.")

    def addLibraryPath(self, index):
        row = index.row()
        self.selectRow(row)
        self.fileDialog.exec()
        if self.fileDialog.selectedFiles():
            self.selectedDirectory = QDir(self.fileDialog.selectedFiles()[0])
        self.model.insertRow(
            row,
            [
                QStandardItem(self.selectedDirectory.dirName()),
                QStandardItem(self.selectedDirectory.absolutePath()),
            ],
        )


class libraryPathEditorDialog(QDialog):
    def __init__(self, parent, libraryDict: dict):
        super().__init__(parent)
        self.parent = parent
        self.logger = self.parent.logger
        self.libraryDict = libraryDict
        self.setWindowTitle("Library Paths Dialogue")
        self.setMinimumSize(700, 300)
        self.mainLayout = QVBoxLayout()
        self.pathsBox = QGroupBox()
        self.boxLayout = QVBoxLayout()
        self.pathsBox.setLayout(self.boxLayout)
        self.pathsModel = libraryPathsModel(self.libraryDict)
        self.tableView = libraryPathsTableView(self.pathsModel, self.logger)
        self.boxLayout.addWidget(self.tableView)
        self.mainLayout.addWidget(self.pathsBox)
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)


class klayoutLaypImportDialogue(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.logger = self.parent.logger

        self.setWindowTitle("KLayout Layout Layers File Importer")
        self.setMinimumSize(500, 250)
        mainLayout = QVBoxLayout()
        fileBox = QGroupBox("Import KLayout Layer Properties File")
        fileBoxLayout = QVBoxLayout()
        inputFileDialogLayout = QHBoxLayout()
        inputFileDialogLayout.addWidget(edf.boldLabel("Layer Properties File:"))
        self.laypFileEdit = edf.longLineEdit()
        inputFileDialogLayout.addWidget(self.laypFileEdit)
        self.laypFileButton = QPushButton("...")
        self.laypFileButton.clicked.connect(self.onFileButtonClicked)
        inputFileDialogLayout.addWidget(self.laypFileButton)
        fileBoxLayout.addLayout(inputFileDialogLayout)
        fileBoxLayout.addSpacing(20)
        outputFileDialogLayout = QHBoxLayout()
        outputFileDialogLayout.addWidget(edf.boldLabel("Output File Directory:"))
        self.outputFileEdit = edf.longLineEdit()
        outputFileDialogLayout.addWidget(self.outputFileEdit)
        self.outputFileButton = QPushButton("...")
        self.outputFileButton.clicked.connect(self.onDirButtonClicked)
        outputFileDialogLayout.addWidget(self.outputFileButton)
        fileBoxLayout.addLayout(outputFileDialogLayout)
        fileBox.setLayout(fileBoxLayout)
        mainLayout.addWidget(fileBox)
        mainLayout.addSpacing(20)

        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        mainLayout.addWidget(self.buttonBox)
        self.setLayout(mainLayout)

    def onFileButtonClicked(self):
        fileDialog = QFileDialog()
        fileDialog.setNameFilter("Layout Properties files (*.lyp)")
        laypFileName = fileDialog.getOpenFileName(
            self, caption="Select LayoutProperties file.", dir=str(pathlib.Path.cwd()),
            filter="Layout Properties files (*.lyp)")[0]
        if laypFileName:
            self.laypFileEdit.setText(laypFileName)

    def onDirButtonClicked(self):
        dirDialog = QFileDialog()
        dirDialog.setFileMode(QFileDialog.Directory)
        dirDialog.setOption(QFileDialog.ShowDirsOnly)
        dirDialog.exec()
        if dirDialog.selectedFiles():
            self.outputFileEdit.setText(dirDialog.selectedFiles()[0])


class klayoutLaytImportDialogue(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.logger = self.parent.logger

        self.setWindowTitle("KLayout Layout Technology File Importer")
        self.setMinimumSize(500, 100)
        mainLayout = QVBoxLayout()
        fileBox = QGroupBox("Import KLayout Technology File")
        fileDialogLayout = QHBoxLayout()
        fileDialogLayout.addWidget(edf.boldLabel("Technology Properties File:"))
        self.laytFileEdit = edf.longLineEdit()
        fileDialogLayout.addWidget(self.laytFileEdit)
        self.laytFileButton = QPushButton("...")
        self.laytFileButton.clicked.connect(self.onFileButtonClicked)
        fileDialogLayout.addWidget(self.laytFileButton)
        fileBox.setLayout(fileDialogLayout)
        mainLayout.addWidget(fileBox)
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        mainLayout.addWidget(self.buttonBox)
        self.setLayout(mainLayout)

    def onFileButtonClicked(self):
        fileDialog = QFileDialog()
        fileDialog.setNameFilter("Layout Properties files (*.lyt)")
        laytFileName = fileDialog.getOpenFileName(
            self, caption="Select LayoutProperties file.", dir=str(pathlib.Path.cwd()),
            filter="Layout Properties files (*.lyt)")[0]
        if laytFileName:
            self.laytFileEdit.setText(laytFileName)


class xschemSymIimportDialogue(QDialog):
    def __init__(self, parent, model):
        super().__init__(parent)
        self.parent = parent
        self.logger = self.parent.logger
        self.model = model
        # print(self.model)
        self.setWindowTitle("Xschem Symbol File Importer")
        self.setMinimumSize(500, 300)
        mainLayout = QVBoxLayout()
        fileBox = QGroupBox("Import Xschem Symbol Files")
        fileDialogLayout = QHBoxLayout()
        fileDialogLayout.addWidget(edf.boldLabel("Xschem Symbol Files:"))
        self.symFileEdit = edf.longLineEdit()
        fileDialogLayout.addWidget(self.symFileEdit)
        self.symFileButton = QPushButton("...")
        self.symFileButton.clicked.connect(self.onFileButtonClicked)
        fileDialogLayout.addWidget(self.symFileButton)
        fileBox.setLayout(fileDialogLayout)
        mainLayout.addWidget(fileBox)
        libraryBox = QGroupBox('Select Library')
        libraryBoxLayout = QFormLayout()
        self.libNamesCB = QComboBox()
        self.libNamesCB.setModel(self.model)
        self.libNamesCB.setModelColumn(0)
        self.libNamesCB.setCurrentIndex(0)
        libraryBoxLayout.addRow(edf.boldLabel("Library:"), self.libNamesCB)
        libraryBox.setLayout(libraryBoxLayout)
        mainLayout.addWidget(libraryBox)
        parameterBox = QGroupBox('Import Parameters')
        parameterBoxLayout = QFormLayout()
        self.scaleEdit = edf.longLineEdit()
        self.scaleEdit.setText('4')
        parameterBoxLayout.addRow(edf.boldLabel("Scale Factor"), self.scaleEdit)
        parameterBox.setLayout(parameterBoxLayout)
        mainLayout.addWidget(parameterBox)
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        mainLayout.addWidget(self.buttonBox)
        self.setLayout(mainLayout)

    def onFileButtonClicked(self):
        fileDialog = QFileDialog()
        fileDialog.setNameFilter("Xschem Symbol Files (*.sym)")
        fileDialog.setFileMode(QFileDialog.ExistingFiles)
        if fileDialog.exec():
            symFileNames = fileDialog.selectedFiles()
            if symFileNames:
                self.symFileEdit.setText(', '.join(symFileNames))


class fileInfoDialogue(QDialog):
    def __init__(self, filePath: pathlib.Path, parent=None):
        super().__init__(parent)
        self.setWindowTitle("File Information")
        self.setMinimumSize(500, 200)
        layout = QFormLayout(self)
        # Get file stats
        stats = filePath.stat()
        created = datetime.datetime.fromtimestamp(stats.st_ctime)
        modified = datetime.datetime.fromtimestamp(stats.st_mtime)
        size = stats.st_size

        layout.addRow("File:", QLabel(filePath.name))
        layout.addRow("Path:", QLabel(str(filePath.absolute())))
        layout.addRow("Size:", QLabel(f"{size:,} bytes"))
        layout.addRow("Created:", QLabel(created.strftime("%Y-%m-%d %H:%M:%S")))
        layout.addRow("Modified:", QLabel(modified.strftime("%Y-%m-%d %H:%M:%S")))
        layout.addRow("Accessed:", QLabel(
            datetime.datetime.fromtimestamp(stats.st_atime).strftime("%Y-%m-%d %H:%M:%S")))
        layout.addRow("Permissions:", QLabel(oct(stats.st_mode)[-3:]))
        # layout.addRow("Owner:", QLabel(str(stats.st_uid)))
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok)
        self.buttonBox.accepted.connect(self.accept)
        layout.addWidget(self.buttonBox)

        self.setLayout(layout)
